#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type
from ansible.plugins.action import ActionBase
try:
    from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
        AnsibleArgSpecValidator,
    )
except ImportError:
    ANSIBLE_UTILS_IS_INSTALLED = False
else:
    ANSIBLE_UTILS_IS_INSTALLED = True
from ansible.errors import AnsibleActionFail
from ansible_collections.cisco.meraki.plugins.plugin_utils.meraki import (
    MERAKI,
    meraki_argument_spec,
    meraki_compare_equality,
    get_dict_result,
)
from ansible_collections.cisco.meraki.plugins.plugin_utils.exceptions import (
    InconsistentParameters,
)

# Get common arguments specification
argument_spec = meraki_argument_spec()
# Add arguments specific for this module
argument_spec.update(dict(
    state=dict(type="str", default="present", choices=["present"]),
    name=dict(type="str"),
    tags=dict(type="list"),
    lat=dict(type="float"),
    lng=dict(type="float"),
    address=dict(type="str"),
    notes=dict(type="str"),
    moveMapMarker=dict(type="bool"),
    switchProfileId=dict(type="str"),
    floorPlanId=dict(type="str"),
    mac=dict(type="str"),
    serial=dict(type="str"),
    organizationId=dict(type="str"),
))

required_if = [
    ("state", "present", ["name", "organizationId", "serial"], True),
]
required_one_of = []
mutually_exclusive = []
required_together = []


class Devices(object):
    def __init__(self, params, meraki):
        self.meraki = meraki
        self.new_object = dict(
            name=params.get("name"),
            tags=params.get("tags"),
            lat=params.get("lat"),
            lng=params.get("lng"),
            address=params.get("address"),
            notes=params.get("notes"),
            moveMapMarker=params.get("moveMapMarker"),
            switchProfileId=params.get("switchProfileId"),
            floorPlanId=params.get("floorPlanId"),
            mac=params.get("mac"),
            serial=params.get("serial"),
            organization_id=params.get("organizationId"),
        )

    def get_all_params(self, name=None, id=None):
        new_object_params = {}
        if self.new_object.get('perPage') is not None or self.new_object.get('per_page') is not None:
            new_object_params['perPage'] = self.new_object.get('perPage') or \
                self.new_object.get('per_page')
        new_object_params['total_pages'] = -1
        if self.new_object.get('startingAfter') is not None or self.new_object.get('starting_after') is not None:
            new_object_params['startingAfter'] = self.new_object.get('startingAfter') or \
                self.new_object.get('starting_after')
        if self.new_object.get('endingBefore') is not None or self.new_object.get('ending_before') is not None:
            new_object_params['endingBefore'] = self.new_object.get('endingBefore') or \
                self.new_object.get('ending_before')
        if self.new_object.get('configurationUpdatedAfter') is not None or self.new_object.get('configuration_updated_after') is not None:
            new_object_params['configurationUpdatedAfter'] = self.new_object.get('configurationUpdatedAfter') or \
                self.new_object.get('configuration_updated_after')
        if self.new_object.get('networkIds') is not None or self.new_object.get('network_ids') is not None:
            new_object_params['networkIds'] = self.new_object.get('networkIds') or \
                self.new_object.get('network_ids')
        if self.new_object.get('productTypes') is not None or self.new_object.get('product_types') is not None:
            new_object_params['productTypes'] = self.new_object.get('productTypes') or \
                self.new_object.get('product_types')
        if self.new_object.get('tags') is not None or self.new_object.get('tags') is not None:
            new_object_params['tags'] = self.new_object.get('tags')
        if self.new_object.get('tagsFilterType') is not None or self.new_object.get('tags_filter_type') is not None:
            new_object_params['tagsFilterType'] = self.new_object.get('tagsFilterType') or \
                self.new_object.get('tags_filter_type')
        new_object_params['name'] = name or self.new_object.get('name')
        if self.new_object.get('mac') is not None or self.new_object.get('mac') is not None:
            new_object_params['mac'] = self.new_object.get('mac')
        if self.new_object.get('serial') is not None or self.new_object.get('serial') is not None:
            new_object_params['serial'] = self.new_object.get('serial')
        if self.new_object.get('model') is not None or self.new_object.get('model') is not None:
            new_object_params['model'] = self.new_object.get('model')
        if self.new_object.get('macs') is not None or self.new_object.get('macs') is not None:
            new_object_params['macs'] = self.new_object.get('macs')
        if self.new_object.get('serials') is not None or self.new_object.get('serials') is not None:
            new_object_params['serials'] = self.new_object.get('serials')
        if self.new_object.get('sensorMetrics') is not None or self.new_object.get('sensor_metrics') is not None:
            new_object_params['sensorMetrics'] = self.new_object.get('sensorMetrics') or \
                self.new_object.get('sensor_metrics')
        if self.new_object.get('sensorAlertProfileIds') is not None or self.new_object.get('sensor_alert_profile_ids') is not None:
            new_object_params['sensorAlertProfileIds'] = self.new_object.get('sensorAlertProfileIds') or \
                self.new_object.get('sensor_alert_profile_ids')
        if self.new_object.get('models') is not None or self.new_object.get('models') is not None:
            new_object_params['models'] = self.new_object.get('models')
        if self.new_object.get('organizationId') is not None or self.new_object.get('organization_id') is not None:
            new_object_params['organizationId'] = self.new_object.get('organizationId') or \
                self.new_object.get('organization_id')
        return new_object_params

    def get_params_by_id(self, name=None, id=None):
        new_object_params = {}
        if self.new_object.get('serial') is not None or self.new_object.get('serial') is not None:
            new_object_params['serial'] = self.new_object.get('serial')
        return new_object_params

    def update_by_id_params(self):
        new_object_params = {}
        if self.new_object.get('name') is not None or self.new_object.get('name') is not None:
            new_object_params['name'] = self.new_object.get('name') or \
                self.new_object.get('name')
        if self.new_object.get('tags') is not None or self.new_object.get('tags') is not None:
            new_object_params['tags'] = self.new_object.get('tags') or \
                self.new_object.get('tags')
        if self.new_object.get('lat') is not None or self.new_object.get('lat') is not None:
            new_object_params['lat'] = self.new_object.get('lat') or \
                self.new_object.get('lat')
        if self.new_object.get('lng') is not None or self.new_object.get('lng') is not None:
            new_object_params['lng'] = self.new_object.get('lng') or \
                self.new_object.get('lng')
        if self.new_object.get('address') is not None or self.new_object.get('address') is not None:
            new_object_params['address'] = self.new_object.get('address') or \
                self.new_object.get('address')
        if self.new_object.get('notes') is not None or self.new_object.get('notes') is not None:
            new_object_params['notes'] = self.new_object.get('notes') or \
                self.new_object.get('notes')
        if self.new_object.get('moveMapMarker') is not None or self.new_object.get('move_map_marker') is not None:
            new_object_params['moveMapMarker'] = self.new_object.get(
                'moveMapMarker')
        if self.new_object.get('switchProfileId') is not None or self.new_object.get('switch_profile_id') is not None:
            new_object_params['switchProfileId'] = self.new_object.get('switchProfileId') or \
                self.new_object.get('switch_profile_id')
        if self.new_object.get('floorPlanId') is not None or self.new_object.get('floor_plan_id') is not None:
            new_object_params['floorPlanId'] = self.new_object.get('floorPlanId') or \
                self.new_object.get('floor_plan_id')
        if self.new_object.get('mac') is not None or self.new_object.get('mac') is not None:
            new_object_params['mac'] = self.new_object.get('mac') or \
                self.new_object.get('mac')
        if self.new_object.get('serial') is not None or self.new_object.get('serial') is not None:
            new_object_params['serial'] = self.new_object.get('serial') or \
                self.new_object.get('serial')
        return new_object_params

    def get_object_by_name(self, name):
        result = None
        # NOTE: Does not have a get by name method, using get all
        try:
            items = self.meraki.exec_meraki(
                family="organizations",
                function="getOrganizationDevices",
                params=self.get_all_params(name=name),
            )
            if isinstance(items, dict):
                if 'response' in items:
                    items = items.get('response')
            result = get_dict_result(items, 'name', name)
            if result is None:
                result = items
        except Exception as e:
            print("Error: ", e)
            result = None
        return result

    def get_object_by_id(self, id):
        result = None
        try:
            items = self.meraki.exec_meraki(
                family="devices",
                function="getDevice",
                params=self.get_params_by_id()
            )
            if isinstance(items, dict):
                if 'response' in items:
                    items = items.get('response')
            result = items
        except Exception as e:
            print("Error: ", e)
            result = None
        return result

    def exists(self):
        prev_obj = None
        id_exists = False
        name_exists = False
        o_id = self.new_object.get("id")
        o_id = o_id or self.new_object.get(
            "serial") or self.new_object.get("serial")
        name = self.new_object.get("name")
        orgID = self.new_object.get("organizationId") or self.new_object.get(
            "organization_id")
        print(name)
        print(orgID)
        if o_id:
            prev_obj = self.get_object_by_id(o_id)
            id_exists = prev_obj is not None and isinstance(prev_obj, dict)
        if not id_exists and name and orgID:
            prev_obj = self.get_object_by_name(name)
            name_exists = prev_obj is not None and isinstance(prev_obj, dict)
        if name_exists:
            _id = prev_obj.get("id")
            _id = _id or prev_obj.get("serial")
            if id_exists and name_exists and o_id != _id:
                raise InconsistentParameters(
                    "The 'id' and 'name' params don't refer to the same object")
            if _id:
                self.new_object.update(dict(id=_id))
                self.new_object.update(dict(serial=_id))
            if _id:
                prev_obj = self.get_object_by_id(_id)
        it_exists = prev_obj is not None and isinstance(prev_obj, dict)
        return (it_exists, prev_obj)

    def requires_update(self, current_obj):
        requested_obj = self.new_object

        obj_params = [
            ("name", "name"),
            ("tags", "tags"),
            ("lat", "lat"),
            ("lng", "lng"),
            ("address", "address"),
            ("notes", "notes"),
            ("moveMapMarker", "moveMapMarker"),
            ("switchProfileId", "switchProfileId"),
            ("floorPlanId", "floorPlanId"),
            ("mac", "mac"),
            ("serial", "serial"),
            ("organizationId", "organizationId"),
        ]
        # Method 1. Params present in request (Ansible) obj are the same as the current (ISE) params
        # If any does not have eq params, it requires update
        return any(not meraki_compare_equality(current_obj.get(meraki_param),
                                               requested_obj.get(ansible_param))
                   for (meraki_param, ansible_param) in obj_params)

    def update(self):
        id = self.new_object.get("id")
        id = id or self.new_object.get("serial")
        name = self.new_object.get("name")
        result = None
        if not id:
            prev_obj_name = self.get_object_by_name(name)
            id_ = None
            if prev_obj_name:
                id_ = prev_obj_name.get("id")
                id_ = id_ or prev_obj_name.get("serial")
            if id_:
                self.new_object.update(dict(serial=id_))
        result = self.meraki.exec_meraki(
            family="devices",
            function="updateDevice",
            params=self.update_by_id_params(),
            op_modifies=True,
        )
        return result


class ActionModule(ActionBase):
    def __init__(self, *args, **kwargs):
        if not ANSIBLE_UTILS_IS_INSTALLED:
            raise AnsibleActionFail(
                "ansible.utils is not installed. Execute 'ansible-galaxy collection install ansible.utils'")
        super(ActionModule, self).__init__(*args, **kwargs)
        self._supports_async = False
        self._supports_check_mode = False
        self._result = None

    # Checks the supplied parameters against the argument spec for this module
    def _check_argspec(self):
        aav = AnsibleArgSpecValidator(
            data=self._task.args,
            schema=dict(argument_spec=argument_spec),
            schema_format="argspec",
            schema_conditionals=dict(
                required_if=required_if,
                required_one_of=required_one_of,
                mutually_exclusive=mutually_exclusive,
                required_together=required_together,
            ),
            name=self._task.action,
        )
        valid, errors, self._task.args = aav.validate()
        if not valid:
            raise AnsibleActionFail(errors)

    def run(self, tmp=None, task_vars=None):
        self._task.diff = False
        self._result = super(ActionModule, self).run(tmp, task_vars)
        self._result["changed"] = False
        self._check_argspec()

        meraki = MERAKI(self._task.args)
        obj = Devices(self._task.args, meraki)

        state = self._task.args.get("state")

        response = None
        if state == "present":
            (obj_exists, prev_obj) = obj.exists()
            if obj_exists:
                if obj.requires_update(prev_obj):
                    response = obj.update()
                    meraki.object_updated()
                else:
                    response = prev_obj
                    meraki.object_already_present()
            else:
                meraki.fail_json(
                    "Object does not exists, plugin only has update")

        self._result.update(dict(meraki_response=response))
        self._result.update(meraki.exit_json())
        return self._result



# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Ntp_global parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


class Ntp_globalTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(Ntp_globalTemplate, self).__init__(lines=lines, tmplt=self, module=module)

    # fmt: off
    PARSERS = [
        {
            "name": "peer",
            "getval": re.compile(
                r"""
                ^ntp\saccess-group
                (\s(?P<ipv4>ipv4))?
                (\s(?P<ipv6>ipv6))?
                \speer
                \s(?P<access_list>\S+)
                (\s(?P<kod>kod))?
                $""", re.VERBOSE,
            ),
            "setval": "ntp access-group"
                      "{{ ' ipv4' if ipv4 is defined else '' }}"
                      "{{ ' ipv6' if ipv6 is defined else '' }}"
                      " peer "
                      "{{ access_list }}"
                      "{{ ' kod' if kod|d(False) else '' }}",
            "result": {
                "access_group": {
                    "peer": [
                        {
                            "access_list": "{{ access_list }}",
                            "kod": "{{ not not kod }}",
                            "ipv4": "{{ not not ipv4 }}",
                            "ipv6": "{{ not not ipv6 }}",
                        },
                    ],
                },
            },
        },
        {
            "name": "query_only",
            "getval": re.compile(
                r"""
                ^ntp\saccess-group
                (\s(?P<ipv4>ipv4))?
                (\s(?P<ipv6>ipv6))?
                \squery-only
                \s(?P<access_list>\S+)
                (\s(?P<kod>kod))?
                $""", re.VERBOSE,
            ),
            "setval": "ntp access-group"
                      "{{ ' ipv4' if ipv4 is defined else '' }}"
                      "{{ ' ipv6' if ipv6 is defined else '' }}"
                      " query-only "
                      "{{ access_list }}"
                      "{{ ' kod' if kod|d(False) else '' }}",
            "result": {
                "access_group": {
                    "query_only": [
                        {
                            "access_list": "{{ access_list }}",
                            "kod": "{{ not not kod }}",
                            "ipv4": "{{ not not ipv4 }}",
                            "ipv6": "{{ not not ipv6 }}",
                        },
                    ],
                },
            },
        },
        {
            "name": "serve",
            "getval": re.compile(
                r"""
                ^ntp\saccess-group
                (\s(?P<ipv4>ipv4))?
                (\s(?P<ipv6>ipv6))?
                \sserve
                \s(?P<access_list>\S+)
                (\s(?P<kod>kod))?
                $""", re.VERBOSE,
            ),
            "setval": "ntp access-group"
                      "{{ ' ipv4' if ipv4 is defined else '' }}"
                      "{{ ' ipv6' if ipv6 is defined else '' }}"
                      " serve "
                      "{{ access_list }}"
                      "{{ ' kod' if kod|d(False) else '' }}",
            "result": {
                "access_group": {
                    "serve": [
                        {
                            "access_list": "{{ access_list }}",
                            "kod": "{{ not not kod }}",
                            "ipv4": "{{ not not ipv4 }}",
                            "ipv6": "{{ not not ipv6 }}",
                        },
                    ],
                },
            },
        },
        {
            "name": "serve_only",
            "getval": re.compile(
                r"""
                ^ntp\saccess-group
                (\s(?P<ipv4>ipv4))?
                (\s(?P<ipv6>ipv6))?
                \sserve-only
                \s(?P<access_list>\S+)
                (\s(?P<kod>kod))?
                $""", re.VERBOSE,
            ),
            "setval": "ntp access-group"
                      "{{ ' ipv4' if ipv4 is defined else '' }}"
                      "{{ ' ipv6' if ipv6 is defined else '' }}"
                      " serve-only "
                      "{{ access_list }}"
                      "{{ ' kod' if kod|d(False) else '' }}",
            "result": {
                "access_group": {
                    "serve_only": [
                        {
                            "access_list": "{{ access_list }}",
                            "kod": "{{ not not kod }}",
                            "ipv4": "{{ not not ipv4 }}",
                            "ipv6": "{{ not not ipv6 }}",
                        },
                    ],
                },
            },
        },
        {
            "name": "allow.control.rate_limit",
            "getval": re.compile(
                r"""
                ^ntp\sallow\smode\scontrol\s(?P<rate_limit>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "ntp allow mode control {{ allow.control.rate_limit }}",
            "result": {
                "allow": {
                    "control": {
                        "rate_limit": "{{ rate_limit }}",
                    },
                },
            },
        },
        {
            "name": "allow.private",
            "getval": re.compile(
                r"""
                ^ntp\sallow\smode\s(?P<private>private)
                $""", re.VERBOSE,
            ),
            "setval": "ntp allow mode private",
            "result": {
                "allow": {
                    "private": "{{ not not private }}",
                },
            },
        },
        {
            "name": "authenticate",
            "getval": re.compile(
                r"""
                ^ntp\s(?P<authenticate>authenticate)
                $""", re.VERBOSE,
            ),
            "setval": "ntp authenticate",
            "result": {
                "authenticate": "{{ not not authenticate }}",
            },
        },
        {
            "name": "authentication_keys",
            "getval": re.compile(
                r"""
                ^ntp\sauthentication-key\s(?P<id>\d+)
                \s(?P<algorithm>\S+)
                \s(?P<key>\S+)
                \s(?P<encryption>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "ntp authentication-key {{ id }} {{ algorithm }} {{ key }} {{ encryption }}",
            "result": {
                "authentication_keys": [
                    {
                        "id": "{{ id }}",
                        "algorithm": "{{ algorithm }}",
                        "key": "{{ key }}",
                        "encryption": "{{ encryption }}",
                    },
                ],
            },
        },
        {
            "name": "broadcast_delay",
            "getval": re.compile(
                r"""
                ^ntp\sbroadcastdelay\s(?P<broadcast_delay>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "ntp broadcastdelay {{ broadcast_delay }}",
            "result": {
                "broadcast_delay": "{{ broadcast_delay }}",
            },
        },
        {
            "name": "clock_period",
            "getval": re.compile(
                r"""
                ^ntp\sclock-period\s(?P<clock_period>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "ntp clock-period {{ clock_period }}",
            "result": {
                "clock_period": "{{ clock_period }}",
            },
        },
        {
            "name": "logging",
            "getval": re.compile(
                r"""
                ^ntp\s(?P<logging>logging)
                $""", re.VERBOSE,
            ),
            "setval": "ntp logging",
            "result": {
                "logging": "{{ not not logging }}",
            },
        },
        {
            "name": "master.enabled",
            "getval": re.compile(
                r"""
                ^ntp\s(?P<master>master)
                $""", re.VERBOSE,
            ),
            "setval": "ntp master",
            "result": {
                "master": {
                    "enabled": "{{ not not master }}",
                },
            },
        },
        {
            "name": "master.stratum",
            "getval": re.compile(
                r"""
                ^ntp\smaster\s(?P<stratum>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "ntp master {{ master.stratum }}",
            "result": {
                "master": {
                    "stratum": "{{ stratum }}",
                },
            },
        },
        {
            "name": "max_associations",
            "getval": re.compile(
                r"""
                ^ntp\smax-associations\s(?P<max_associations>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "ntp max-associations {{ max_associations }}",
            "result": {
                "max_associations": "{{ max_associations }}",
            },
        },
        {
            "name": "max_distance",
            "getval": re.compile(
                r"""
                ^ntp\smaxdistance\s(?P<max_distance>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "ntp maxdistance {{ max_distance }}",
            "result": {
                "max_distance": "{{ max_distance }}",
            },
        },
        {
            "name": "min_distance",
            "getval": re.compile(
                r"""
                ^ntp\smindistance\s(?P<min_distance>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "ntp mindistance {{ min_distance }}",
            "result": {
                "min_distance": "{{ min_distance }}",
            },
        },
        {
            "name": "orphan",
            "getval": re.compile(
                r"""
                ^ntp\sorphan\s(?P<orphan>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "ntp orphan {{ orphan }}",
            "result": {
                "orphan": "{{ orphan }}",
            },
        },
        {
            "name": "panic_update",
            "getval": re.compile(
                r"""
                ^ntp\spanic\s(?P<update>update)
                $""", re.VERBOSE,
            ),
            "setval": "ntp panic update",
            "result": {
                "panic_update": "{{ not not update }}",
            },
        },
        {
            "name": "passive",
            "getval": re.compile(
                r"""
                ^ntp\s(?P<passive>passive)
                $""", re.VERBOSE,
            ),
            "setval": "ntp passive",
            "result": {
                "passive": "{{ not not passive }}",
            },
        },
        {
            "name": "peers",
            "getval": re.compile(
                r"""
                ^ntp\speer
                (\svrf\s(?P<vrf>\S+))?
                (\s(?P<ipv4>ip))?
                (\s(?P<ipv6>ipv6))?
                \s(?P<peer>\S+)
                (\s(?P<burst>burst))?
                (\s(?P<iburst>iburst))?
                (\skey\s(?P<key>\d+))?
                (\sminpoll\s(?P<minpoll>\d+))?
                (\smaxpoll\s(?P<maxpoll>\d+))?
                (\s(?P<normal_sync>normal-sync))?
                (\s(?P<prefer>prefer))?
                (\ssource\s(?P<source>\S+))?
                (\sversion\s(?P<version>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "ntp peer"
                      "{{ (' vrf ' + vrf) if vrf is defined else '' }}"
                      "{{ ' ip' if use_ipv4|d(False) else ''}}"
                      "{{ ' ipv6' if use_ipv6|d(False) else ''}}"
                      "{{ ( ' '  + peer ) if peer is defined else '' }}"
                      "{{ ' burst ' if burst|d(False) else ''}}"
                      "{{ ' iburst ' if iburst|d(False) else ''}}"
                      "{{ (' key ' + key_id|string) if key_id is defined else '' }}"
                      "{{ (' minpoll ' + minpoll|string) if minpoll is defined else '' }}"
                      "{{ (' maxpoll ' + maxpoll|string) if maxpoll is defined else '' }}"
                      "{{ ' normal-sync ' if normal_sync is defined else ''}}"
                      "{{ ' prefer' if prefer|d(False) else ''}}"
                      "{{ (' source ' + source) if source is defined else '' }}"
                      "{{ (' version ' + version|string) if version is defined else '' }}",
            "result": {
                "peers": [
                    {
                        "peer": "{{ peer }}",
                        "use_ipv4": "{{ not not ipv4 }}",
                        "use_ipv6": "{{ not not ipv6 }}",
                        "vrf": "{{ vrf }}",
                        "burst": "{{ not not burst }}",
                        "iburst": "{{ not not iburst }}",
                        "key_id": "{{ key }}",
                        "minpoll": "{{ minpoll }}",
                        "maxpoll": "{{ maxpoll }}",
                        "normal_sync": "{{ not not normal_sync }}",
                        "prefer": "{{ not not prefer }}",
                        "source": "{{ source }}",
                        "version": "{{ version }}",
                    },
                ],
            },
        },
        {
            "name": "servers",
            "getval": re.compile(
                r"""
                ^ntp\sserver
                (\svrf\s(?P<vrf>\S+))?
                (\s(?P<ipv4>ip))?
                (\s(?P<ipv6>ipv6))?
                \s(?P<server>\S+)
                (\s(?P<burst>burst))?
                (\s(?P<iburst>iburst))?
                (\skey\s(?P<key>\d+))?
                (\sminpoll\s(?P<minpoll>\d+))?
                (\smaxpoll\s(?P<maxpoll>\d+))?
                (\s(?P<normal_sync>normal-sync))?
                (\s(?P<prefer>prefer))?
                (\ssource\s(?P<source>\S+))?
                (\sversion\s(?P<version>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "ntp server"
                      "{{ (' vrf ' + vrf) if vrf is defined else '' }}"
                      "{{ ' ip' if use_ipv4|d(False) else ''}}"
                      "{{ ' ipv6' if use_ipv6|d(False) else ''}}"
                      "{{ ( ' '  + server ) if server is defined else '' }}"
                      "{{ ' burst ' if burst|d(False) else ''}}"
                      "{{ ' iburst ' if iburst|d(False) else ''}}"
                      "{{ (' key ' + key_id|string) if key_id is defined else '' }}"
                      "{{ (' minpoll ' + minpoll|string) if minpoll is defined else '' }}"
                      "{{ (' maxpoll ' + maxpoll|string) if maxpoll is defined else '' }}"
                      "{{ ' normal-sync ' if normal_sync is defined else ''}}"
                      "{{ ' prefer' if prefer|d(False) else ''}}"
                      "{{ (' source ' + source) if source is defined else '' }}"
                      "{{ (' version ' + version|string) if version is defined else '' }}",
            "result": {
                "servers": [
                    {
                        "server": "{{ server }}",
                        "use_ipv4": "{{ not not ipv4 }}",
                        "use_ipv6": "{{ not not ipv6 }}",
                        "vrf": "{{ vrf }}",
                        "burst": "{{ not not burst }}",
                        "iburst": "{{ not not iburst }}",
                        "key_id": "{{ key }}",
                        "minpoll": "{{ minpoll }}",
                        "maxpoll": "{{ maxpoll }}",
                        "normal_sync": "{{ not not normal_sync }}",
                        "prefer": "{{ not not prefer }}",
                        "source": "{{ source }}",
                        "version": "{{ version }}",
                    },
                ],
            },
        },
        {
            "name": "source",
            "getval": re.compile(
                r"""
                ^ntp\ssource\s(?P<source>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "ntp source {{ source }}",
            "result": {
                "source": "{{ source }}",
            },
        },
        {
            "name": "trusted_keys",
            "getval": re.compile(
                r"""
                ^ntp\strusted-key
                \s((?P<range_start>\d+))
                (\s\-\s)?
                ((?P<range_end>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "ntp trusted-key {{ range_start }}"
                      "{{ (' - ' + range_end|string) if range_end is defined else '' }}",
            "result": {
                "trusted_keys": [
                    {
                        "range_start": "{{ range_start }}",
                        "range_end": "{{ range_end }}",
                    },
                ],
            },
        },
        {
            "name": "update_calendar",
            "getval": re.compile(
                r"""
                ^ntp\s(?P<update_calendar>update-calendar)
                $""", re.VERBOSE,
            ),
            "setval": "ntp update-calendar",
            "result": {
                "update_calendar": "{{ not not update_calendar }}",
            },
        },
    ]
    # fmt: on



# -*- coding: utf-8 -*-
# Copyright (c) 2018, Stefan Heitmueller <stefan.heitmueller@gmx.com>
# Copyright (c) 2018 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

DOCUMENTATION = '''
    name: gitlab_runners
    author:
      - Stefan Heitmüller (@morph027) <stefan.heitmueller@gmx.com>
    short_description: Ansible dynamic inventory plugin for GitLab runners.
    requirements:
        - python-gitlab > 1.8.0
    extends_documentation_fragment:
        - constructed
    description:
        - Reads inventories from the GitLab API.
        - Uses a YAML configuration file gitlab_runners.[yml|yaml].
    options:
        plugin:
            description: The name of this plugin, it should always be set to 'gitlab_runners' for this plugin to recognize it as it's own.
            type: str
            required: true
            choices:
              - gitlab_runners
              - community.general.gitlab_runners
        server_url:
            description: The URL of the GitLab server, with protocol (i.e. http or https).
            env:
              - name: GITLAB_SERVER_URL
                version_added: 1.0.0
            type: str
            required: true
        api_token:
            description: GitLab token for logging in.
            env:
              - name: GITLAB_API_TOKEN
                version_added: 1.0.0
            type: str
            aliases:
              - private_token
              - access_token
        filter:
            description: filter runners from GitLab API
            env:
              - name: GITLAB_FILTER
                version_added: 1.0.0
            type: str
            choices: ['active', 'paused', 'online', 'specific', 'shared']
        verbose_output:
            description: Toggle to (not) include all available nodes metadata
            type: bool
            default: true
'''

EXAMPLES = '''
# gitlab_runners.yml
plugin: community.general.gitlab_runners
host: https://gitlab.com

# Example using constructed features to create groups and set ansible_host
plugin: community.general.gitlab_runners
host: https://gitlab.com
strict: false
keyed_groups:
  # add e.g. amd64 hosts to an arch_amd64 group
  - prefix: arch
    key: 'architecture'
  # add e.g. linux hosts to an os_linux group
  - prefix: os
    key: 'platform'
  # create a group per runner tag
  # e.g. a runner tagged w/ "production" ends up in group "label_production"
  # hint: labels containing special characters will be converted to safe names
  - key: 'tag_list'
    prefix: tag
'''

from ansible.errors import AnsibleError, AnsibleParserError
from ansible.module_utils.common.text.converters import to_native
from ansible.plugins.inventory import BaseInventoryPlugin, Constructable
from ansible.utils.unsafe_proxy import wrap_var as make_unsafe

try:
    import gitlab
    HAS_GITLAB = True
except ImportError:
    HAS_GITLAB = False


class InventoryModule(BaseInventoryPlugin, Constructable):
    ''' Host inventory parser for ansible using GitLab API as source. '''

    NAME = 'community.general.gitlab_runners'

    def _populate(self):
        gl = gitlab.Gitlab(self.get_option('server_url'), private_token=self.get_option('api_token'))
        self.inventory.add_group('gitlab_runners')
        try:
            if self.get_option('filter'):
                runners = gl.runners.all(scope=self.get_option('filter'))
            else:
                runners = gl.runners.all()
            for runner in runners:
                host = make_unsafe(str(runner['id']))
                ip_address = runner['ip_address']
                host_attrs = make_unsafe(vars(gl.runners.get(runner['id']))['_attrs'])
                self.inventory.add_host(host, group='gitlab_runners')
                self.inventory.set_variable(host, 'ansible_host', make_unsafe(ip_address))
                if self.get_option('verbose_output', True):
                    self.inventory.set_variable(host, 'gitlab_runner_attributes', host_attrs)

                # Use constructed if applicable
                strict = self.get_option('strict')
                # Composed variables
                self._set_composite_vars(self.get_option('compose'), host_attrs, host, strict=strict)
                # Complex groups based on jinja2 conditionals, hosts that meet the conditional are added to group
                self._add_host_to_composed_groups(self.get_option('groups'), host_attrs, host, strict=strict)
                # Create groups based on variable values and add the corresponding hosts to it
                self._add_host_to_keyed_groups(self.get_option('keyed_groups'), host_attrs, host, strict=strict)
        except Exception as e:
            raise AnsibleParserError('Unable to fetch hosts from GitLab API, this was the original exception: %s' % to_native(e))

    def verify_file(self, path):
        """Return the possibly of a file being consumable by this plugin."""
        return (
            super(InventoryModule, self).verify_file(path) and
            path.endswith(("gitlab_runners.yaml", "gitlab_runners.yml")))

    def parse(self, inventory, loader, path, cache=True):
        if not HAS_GITLAB:
            raise AnsibleError('The GitLab runners dynamic inventory plugin requires python-gitlab: https://python-gitlab.readthedocs.io/en/stable/')
        super(InventoryModule, self).parse(inventory, loader, path, cache)
        self._read_config_data(path)
        self._populate()



# Copyright (c) 2019-2020, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


from ansible.errors import AnsibleConnectionFailure
from ansible.utils.display import Display

from ansible_collections.community.docker.plugins.module_utils.common_api import (
    AnsibleDockerClientBase,
)

from ansible_collections.community.docker.plugins.module_utils.util import (
    DOCKER_COMMON_ARGS,
)


class AnsibleDockerClient(AnsibleDockerClientBase):
    def __init__(self, plugin, min_docker_api_version=None):
        self.plugin = plugin
        self.display = Display()
        super(AnsibleDockerClient, self).__init__(
            min_docker_api_version=min_docker_api_version)

    def fail(self, msg, **kwargs):
        if kwargs:
            msg += '\nContext:\n' + '\n'.join('  {0} = {1!r}'.format(k, v) for (k, v) in kwargs.items())
        raise AnsibleConnectionFailure(msg)

    def deprecate(self, msg, version=None, date=None, collection_name=None):
        self.display.deprecated(msg, version=version, date=date, collection_name=collection_name)

    def _get_params(self):
        return dict([
            (option, self.plugin.get_option(option))
            for option in DOCKER_COMMON_ARGS
        ])



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: ucs_ip_pool
short_description: Configures IP address pools on Cisco UCS Manager
description:
- Configures IP address pools and blocks of IP addresses on Cisco UCS Manager.
extends_documentation_fragment: cisco.ucs.ucs
options:
  state:
    description:
    - If C(present), will verify IP pool is present and will create if needed.
    - If C(absent), will verify IP pool is absent and will delete if needed.
    choices: [present, absent]
    default: present
  name:
    description:
    - The name of the IP address pool.
    - This name can be between 1 and 32 alphanumeric characters.
    - "You cannot use spaces or any special characters other than - (hyphen), \"_\" (underscore), : (colon), and . (period)."
    - You cannot change this name after the IP address pool is created.
    required: yes
  description:
    description:
    - The user-defined description of the IP address pool.
    - Enter up to 256 characters.
    - "You can use any characters or spaces except the following:"
    - "` (accent mark), \ (backslash), ^ (carat), \" (double quote), = (equal sign), > (greater than), < (less than), or ' (single quote)."
    aliases: [ descr ]
  order:
    description:
    - The Assignment Order field.
    - "This can be one of the following:"
    - "default - Cisco UCS Manager selects a random identity from the pool."
    - "sequential - Cisco UCS Manager selects the lowest available identity from the pool."
    choices: [default, sequential]
    default: default
  ip_blocks:
    description:
    - List of IPv4 blocks used by the IP Pool.
    suboptions:
      first_addr:
        description:
        - The first IPv4 address in the IPv4 addresses block.
        - This is the From field in the UCS Manager Add IPv4 Blocks menu.
      last_addr:
        description:
        - The last IPv4 address in the IPv4 addresses block.
        - This is the To field in the UCS Manager Add IPv4 Blocks menu.
      subnet_mask:
        description:
        - The subnet mask associated with the IPv4 addresses in the block.
        default: 255.255.255.0
      default_gw:
        description:
        - The default gateway associated with the IPv4 addresses in the block.
        default: 0.0.0.0
      primary_dns:
        description:
        - The primary DNS server that this block of IPv4 addresses should access.
        default: 0.0.0.0
      secondary_dns:
        description:
        - The secondary DNS server that this block of IPv4 addresses should access.
        default: 0.0.0.0
  ipv6_blocks:
    description:
    - List of IPv6 blocks used by the IP Pool.
    suboptions:
      ipv6_first_addr:
        description:
        - The first IPv6 address in the IPv6 addresses block.
        - This is the From field in the UCS Manager Add IPv6 Blocks menu.
      ipv6_last_addr:
        description:
        - The last IPv6 address in the IPv6 addresses block.
        - This is the To field in the UCS Manager Add IPv6 Blocks menu.
      ipv6_prefix:
        description:
        - The network address prefix associated with the IPv6 addresses in the block.
        default: '64'
      ipv6_default_gw:
        description:
        - The default gateway associated with the IPv6 addresses in the block.
        default: '::'
      ipv6_primary_dns:
        description:
        - The primary DNS server that this block of IPv6 addresses should access.
        default: '::'
      ipv6_secondary_dns:
        description:
        - The secondary DNS server that this block of IPv6 addresses should access.
        default: '::'
  org_dn:
    description:
    - Org dn (distinguished name)
    default: org-root
requirements:
- ucsmsdk
author:
  - Brett Johnson (@sdbrett)
  - David Soper (@dsoper2)
  - John McDonough (@movinalot)
  - CiscoUcs (@CiscoUcs)
version_added: '2.5'
'''

EXAMPLES = r'''
- name: Configure IPv4 and IPv6 address pool
  cisco.ucs.ucs_ip_pool:
    hostname: "{{ ucs_hostname }}"
    username: "{{ ucs_username }}"
    password: "{{ ucs_password }}"
    name: ip-pool-01
    org_dn: org-root/org-level1
    ipv4_blocks:
    - first_addr: 192.168.10.1
      last_addr: 192.168.10.20
      subnet_mask: 255.255.255.128
      default_gw: 192.168.10.2
    - first_addr: 192.168.11.1
      last_addr: 192.168.11.20
      subnet_mask: 255.255.255.128
      default_gw: 192.168.11.2
    ipv6_blocks:
    - ipv6_first_addr: fe80::1cae:7992:d7a1:ed07
      ipv6_last_addr: fe80::1cae:7992:d7a1:edfe
      ipv6_default_gw: fe80::1cae:7992:d7a1:ecff
    - ipv6_first_addr: fe80::1cae:7992:d7a1:ec07
      ipv6_last_addr: fe80::1cae:7992:d7a1:ecfe
      ipv6_default_gw: fe80::1cae:7992:d7a1:ecff

- name: Delete IPv4 and IPv6 address pool blocks
  cisco.ucs.ucs_ip_pool:
    hostname: "{{ ucs_hostname }}"
    username: "{{ ucs_username }}"
    password: "{{ ucs_password }}"
    name: ip-pool-01
    org_dn: org-root/org-level1
    ipv4_blocks:
    - first_addr: 192.168.10.1
      last_addr: 192.168.10.20
      state: absent
    ipv6_blocks:
    - ipv6_first_addr: fe80::1cae:7992:d7a1:ec07
      ipv6_last_addr: fe80::1cae:7992:d7a1:ecfe
      state: absent

- name: Remove IPv4 and IPv6 address pool
  cisco.ucs.ucs_ip_pool:
    hostname: "{{ ucs_hostname }}"
    username: "{{ ucs_username }}"
    password: "{{ ucs_password }}"
    name: ip-pool-01
    state: absent
'''

RETURN = r'''
#
'''


def update_ip_pool(ucs, module):
    from ucsmsdk.mometa.ippool.IppoolPool import IppoolPool

    mo = IppoolPool(
        parent_mo_or_dn=module.params['org_dn'],
        name=module.params['name'],
        descr=module.params['descr'],
        assignment_order=module.params['order'],
    )
    ucs.login_handle.add_mo(mo, True)
    ucs.login_handle.commit()

    return mo


def match_existing_ipv4_block(ucs, dn, ipv4_block):
    # ipv4 block specified, check properties
    mo_1 = get_ip_block(ucs, dn, ipv4_block['first_addr'], ipv4_block['last_addr'], 'v4')
    if not mo_1:
        if ipv4_block['state'] == 'absent':
            return True
        return False
    else:
        if ipv4_block['state'] == 'absent':
            return False
        kwargs = dict(subnet=ipv4_block['subnet_mask'])
        kwargs['def_gw'] = ipv4_block['default_gw']
        kwargs['prim_dns'] = ipv4_block['primary_dns']
        kwargs['sec_dns'] = ipv4_block['secondary_dns']
        return mo_1.check_prop_match(**kwargs)


def match_existing_ipv6_block(ucs, dn, ipv6_block):
    # ipv6 block specified, check properties
    mo_1 = get_ip_block(ucs, dn, ipv6_block['ipv6_first_addr'], ipv6_block['ipv6_last_addr'], 'v6')
    if not mo_1:
        if ipv6_block['state'] == 'absent':
            return True
        return False
    else:
        if ipv6_block['state'] == 'absent':
            return False
        kwargs = dict(prefix=ipv6_block['ipv6_prefix'])
        kwargs['def_gw'] = ipv6_block['ipv6_default_gw']
        kwargs['prim_dns'] = ipv6_block['ipv6_primary_dns']
        kwargs['sec_dns'] = ipv6_block['ipv6_secondary_dns']
        return mo_1.check_prop_match(**kwargs)


def remove_ip_block(ucs, dn, ip_block, ip_version):
    if ip_version == 'v6':
        first_addr = ip_block['ipv6_first_addr']
        last_addr = ip_block['ipv6_last_addr']
    else:
        first_addr = ip_block['first_addr']
        last_addr = ip_block['last_addr']

    mo_1 = get_ip_block(ucs, dn, first_addr, last_addr, ip_version)
    if mo_1:
        ucs.login_handle.remove_mo(mo_1)
        ucs.login_handle.commit()


def update_ip_block(ucs, mo, ip_block, ip_version):

    remove_ip_block(ucs, mo.dn, ip_block, ip_version)
    if not ip_block['state'] == 'absent':
        if ip_version == 'v6':
            from ucsmsdk.mometa.ippool.IppoolIpV6Block import IppoolIpV6Block
            IppoolIpV6Block(
                parent_mo_or_dn=mo,
                to=ip_block['ipv6_last_addr'],
                r_from=ip_block['ipv6_first_addr'],
                prefix=ip_block['ipv6_prefix'],
                def_gw=ip_block['ipv6_default_gw'],
                prim_dns=ip_block['ipv6_primary_dns'],
                sec_dns=ip_block['ipv6_secondary_dns']
            )
            ucs.login_handle.add_mo(mo, True)
            ucs.login_handle.commit()
        else:
            from ucsmsdk.mometa.ippool.IppoolBlock import IppoolBlock
            IppoolBlock(
                parent_mo_or_dn=mo,
                to=ip_block['last_addr'],
                r_from=ip_block['first_addr'],
                subnet=ip_block['subnet_mask'],
                def_gw=ip_block['default_gw'],
                prim_dns=ip_block['primary_dns'],
                sec_dns=ip_block['secondary_dns']
            )
            ucs.login_handle.add_mo(mo, True)
            ucs.login_handle.commit()


def get_ip_block(ucs, pool_dn, first_addr, last_addr, ip_version):
    if ip_version == 'v6':
        dn_type = '/v6block-'
    else:
        dn_type = '/block-'

    block_dn = pool_dn + dn_type + first_addr + '-' + last_addr
    return ucs.login_handle.query_dn(block_dn)


def main():
    from ansible.module_utils.basic import AnsibleModule
    from ansible_collections.cisco.ucs.plugins.module_utils.ucs import UCSModule, ucs_argument_spec

    ipv4_configuration_spec = dict(
        first_addr=dict(type='str'),
        last_addr=dict(type='str'),
        subnet_mask=dict(type='str', default='255.255.255.0'),
        default_gw=dict(type='str', default='0.0.0.0'),
        primary_dns=dict(type='str', default='0.0.0.0'),
        secondary_dns=dict(type='str', default='0.0.0.0'),
        state=dict(type='str', default='present', choices=['present', 'absent']),
    )
    ipv6_configuration_spec = dict(
        ipv6_first_addr=dict(type='str'),
        ipv6_last_addr=dict(type='str'),
        ipv6_prefix=dict(type='str', default='64'),
        ipv6_default_gw=dict(type='str', default='::'),
        ipv6_primary_dns=dict(type='str', default='::'),
        ipv6_secondary_dns=dict(type='str', default='::'),
        state=dict(type='str', default='present', choices=['present', 'absent']),
    )

    argument_spec = ucs_argument_spec
    argument_spec.update(
        org_dn=dict(type='str', default='org-root'),
        name=dict(type='str', required=True),
        descr=dict(type='str', default='', aliases=['description']),
        order=dict(type='str', default='default', choices=['default', 'sequential']),
        first_addr=dict(type='str'),
        last_addr=dict(type='str'),
        subnet_mask=dict(type='str', default='255.255.255.0'),
        default_gw=dict(type='str', default='0.0.0.0'),
        primary_dns=dict(type='str', default='0.0.0.0'),
        secondary_dns=dict(type='str', default='0.0.0.0'),
        ipv6_first_addr=dict(type='str'),
        ipv6_last_addr=dict(type='str'),
        ipv6_prefix=dict(type='str', default='64'),
        ipv6_default_gw=dict(type='str', default='::'),
        ipv6_primary_dns=dict(type='str', default='::'),
        ipv6_secondary_dns=dict(type='str', default='::'),
        state=dict(type='str', default='present', choices=['present', 'absent']),
        ipv4_blocks=dict(type='list', default=None, elements='dict', options=ipv4_configuration_spec),
        ipv6_blocks=dict(type='list', default=None, elements='dict', options=ipv6_configuration_spec),
    )

    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )
    # UCSModule verifies ucsmsdk is present and exits on failure.  Imports are below ucs object creation.
    ucs = UCSModule(module)

    err = False

    from ucsmsdk.mometa.ippool.IppoolBlock import IppoolBlock
    from ucsmsdk.mometa.ippool.IppoolIpV6Block import IppoolIpV6Block

    changed = False
    try:
        mo_exists = False
        ipv4_props_match = True
        ipv6_props_match = True
        # dn is <org_dn>/ip-pool-<name>
        dn = module.params['org_dn'] + '/ip-pool-' + module.params['name']

        mo = ucs.login_handle.query_dn(dn)
        if mo:
            mo_exists = True
        if module.params['state'] == 'absent':
            if mo_exists:
                if not module.check_mode:
                    ucs.login_handle.remove_mo(mo)
                    ucs.login_handle.commit()
                changed = True
        else:
            if not mo_exists:
                if not module.check_mode:
                    mo = update_ip_pool(ucs, module)
                changed = True
            if mo_exists:
                # check top-level mo props
                kwargs = dict(assignment_order=module.params['order'])
                kwargs['descr'] = module.params['descr']
                if not mo.check_prop_match(**kwargs):
                    if not module.check_mode:
                        mo = update_ip_pool(ucs, module)
                    changed = True
                    # top-level props match, check next level mo/props
            if module.params['ipv4_blocks']:
                for ipv4_block in module.params['ipv4_blocks']:
                    if not match_existing_ipv4_block(ucs, dn, ipv4_block):
                        if not module.check_mode:
                            update_ip_block(ucs, mo, ipv4_block, 'v4')
                        changed = True
            elif module.params['last_addr'] and module.params['first_addr']:
                # ipv4 block specified, check properties
                mo_1 = get_ip_block(ucs, dn, module.params['first_addr'], module.params['last_addr'], 'v4')
                if mo_1:
                    kwargs = dict(subnet=module.params['subnet_mask'])
                    kwargs['def_gw'] = module.params['default_gw']
                    kwargs['prim_dns'] = module.params['primary_dns']
                    kwargs['sec_dns'] = module.params['secondary_dns']
                    if not mo_1.check_prop_match(**kwargs):
                        # ipv4 block exists and properties match
                        ipv4_props_match = False
                else:
                    ipv4_props_match = False

            # only check ipv6 props if the top-level and ipv4 props matched
            if module.params['ipv6_blocks']:
                for ipv6_block in module.params['ipv6_blocks']:
                    if not match_existing_ipv6_block(ucs, dn, ipv6_block):
                        if not module.check_mode:
                            update_ip_block(ucs, mo, ipv6_block, 'v6')
                        changed = True
            elif module.params['ipv6_last_addr'] and module.params['ipv6_first_addr']:
                # ipv6 block specified, check properties
                block_dn = dn + '/v6block-' + module.params['ipv6_first_addr'].lower() + '-' + module.params[
                    'ipv6_last_addr'].lower()
                mo_1 = ucs.login_handle.query_dn(block_dn)
                if mo_1:
                    kwargs = dict(prefix=module.params['ipv6_prefix'])
                    kwargs['def_gw'] = module.params['ipv6_default_gw']
                    kwargs['prim_dns'] = module.params['ipv6_primary_dns']
                    kwargs['sec_dns'] = module.params['ipv6_secondary_dns']
                    if not mo_1.check_prop_match(**kwargs):
                        # ipv6 block exists and properties match
                        ipv6_props_match = False
                else:
                    ipv6_props_match = False

        if not ipv4_props_match or not ipv6_props_match:
            if not module.check_mode:
                if module.params['last_addr'] and module.params['first_addr']:
                    IppoolBlock(
                        parent_mo_or_dn=mo,
                        to=module.params['last_addr'],
                        r_from=module.params['first_addr'],
                        subnet=module.params['subnet_mask'],
                        def_gw=module.params['default_gw'],
                        prim_dns=module.params['primary_dns'],
                        sec_dns=module.params['secondary_dns'],
                    )

                if module.params['ipv6_last_addr'] and module.params['ipv6_first_addr']:
                    IppoolIpV6Block(
                        parent_mo_or_dn=mo,
                        to=module.params['ipv6_last_addr'],
                        r_from=module.params['ipv6_first_addr'],
                        prefix=module.params['ipv6_prefix'],
                        def_gw=module.params['ipv6_default_gw'],
                        prim_dns=module.params['ipv6_primary_dns'],
                        sec_dns=module.params['ipv6_secondary_dns'],
                    )

            ucs.login_handle.add_mo(mo, True)
            ucs.login_handle.commit()

            changed = True

    except Exception as e:
        err = True
        ucs.result['msg'] = "setup error: %s " % str(e)

    ucs.result['changed'] = changed
    if err:
        module.fail_json(**ucs.result)
    module.exit_json(**ucs.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: networks_policies_by_client_info
short_description: Information module for networks _policies _byclient
description:
- Get all networks _policies _byclient.
- Get policies for all clients with policies.
version_added: '2.16.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
  - cisco.meraki.module_info_pagination
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  networkId:
    description:
    - NetworkId path parameter. Network ID.
    type: str
  perPage:
    description:
    - PerPage query parameter. The number of entries per page returned. Acceptable range is 3 - 1000. Default is 50.
    type: int
  startingAfter:
    description:
    - >
      StartingAfter query parameter. A token used by the server to indicate the start of the page. Often this is a
      timestamp or an ID but it is not limited to those. This parameter should not be defined by client
      applications. The link for the first, last, prev, or next page in the HTTP Link header should define it.
    type: str
  endingBefore:
    description:
    - >
      EndingBefore query parameter. A token used by the server to indicate the end of the page. Often this is a
      timestamp or an ID but it is not limited to those. This parameter should not be defined by client
      applications. The link for the first, last, prev, or next page in the HTTP Link header should define it.
    type: str
  t0:
    description:
    - T0 query parameter. The beginning of the timespan for the data. The maximum lookback period is 31 days from today.
    type: str
  timespan:
    description:
    - >
      Timespan query parameter. The timespan for which the information will be fetched. If specifying timespan, do
      not specify parameter t0. The value must be in seconds and be less than or equal to 31 days. The default is
      1 day.
    type: float
requirements:
- meraki >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco Meraki documentation for networks getNetworkPoliciesByClient
  description: Complete reference of the getNetworkPoliciesByClient API.
  link: https://developer.cisco.com/meraki/api-v1/#!get-network-policies-by-client
notes:
  - SDK Method used are
    networks.Networks.get_network_policies_by_client,

  - Paths used are
    get /networks/{networkId}/policies/byClient,
"""

EXAMPLES = r"""
- name: Get all networks _policies _byclient
  cisco.meraki.networks_policies_by_client_info:
    meraki_api_key: "{{meraki_api_key}}"
    meraki_base_url: "{{meraki_base_url}}"
    meraki_single_request_timeout: "{{meraki_single_request_timeout}}"
    meraki_certificate_path: "{{meraki_certificate_path}}"
    meraki_requests_proxy: "{{meraki_requests_proxy}}"
    meraki_wait_on_rate_limit: "{{meraki_wait_on_rate_limit}}"
    meraki_nginx_429_retry_wait_time: "{{meraki_nginx_429_retry_wait_time}}"
    meraki_action_batch_retry_wait_time: "{{meraki_action_batch_retry_wait_time}}"
    meraki_retry_4xx_error: "{{meraki_retry_4xx_error}}"
    meraki_retry_4xx_error_wait_time: "{{meraki_retry_4xx_error_wait_time}}"
    meraki_maximum_retries: "{{meraki_maximum_retries}}"
    meraki_output_log: "{{meraki_output_log}}"
    meraki_log_file_prefix: "{{meraki_log_file_prefix}}"
    meraki_log_path: "{{meraki_log_path}}"
    meraki_print_console: "{{meraki_print_console}}"
    meraki_suppress_logging: "{{meraki_suppress_logging}}"
    meraki_simulate: "{{meraki_simulate}}"
    meraki_be_geo_id: "{{meraki_be_geo_id}}"
    meraki_use_iterator_for_get_pages: "{{meraki_use_iterator_for_get_pages}}"
    meraki_inherit_logging_config: "{{meraki_inherit_logging_config}}"
    perPage: 0
    startingAfter: string
    endingBefore: string
    t0: string
    timespan: 0
    networkId: string
    total_pages: -1
    direction: next
  register: result

"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: list
  elements: dict
  sample: >
    [
      {
        "name": "string",
        "clientId": "string",
        "assigned": [
          {
            "name": "string",
            "type": "string",
            "groupPolicyId": "string",
            "ssid": [
              {
                "ssidNumber": 0
              }
            ]
          }
        ]
      }
    ]
"""



#!/usr/bin/python
#
# Copyright (c) 2016 Matt Davis, <mdavis@ansible.com>
#                    Chris Houseknecht, <house@redhat.com>
#                    Yuwei ZHou, <yuwzho@microsoft.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_networkinterface

version_added: "0.1.0"

short_description: Manage Azure network interfaces

description:
    - Create, update or delete a network interface.
    - When creating a network interface you must provide the name of an existing virtual network, the name of an existing subnet within the virtual network.
    - A default security group and public IP address will be created automatically.
    - Or you can provide the name of an existing security group and public IP address.
    - See the examples below for more details.

options:
    resource_group:
        description:
            - Name of a resource group where the network interface exists or will be created.
        required: true
        type: str
    name:
        description:
            - Name of the network interface.
        required: true
        type: str
    state:
        description:
            - Assert the state of the network interface. Use C(present) to create or update an interface and
              C(absent) to delete an interface.
        default: present
        type: str
        choices:
            - absent
            - present
    location:
        description:
            - Valid Azure location. Defaults to location of the resource group.
        type: str
    virtual_network:
        description:
            - An existing virtual network with which the network interface will be associated. Required when creating a network interface.
            - It can be the virtual network's name.
            - Make sure your virtual network is in the same resource group as NIC when you give only the name.
            - It can be the virtual network's resource id.
            - It can be a dict which contains I(name) and I(resource_group) of the virtual network.
            - Required when creating.
        type: raw
        aliases:
            - virtual_network_name
    subnet_name:
        description:
            - Name of an existing subnet within the specified virtual network. Required when creating a network interface.
            - Use the C(virtual_network)'s resource group.
            - Required when creating.
        aliases:
            - subnet
        type: str
    os_type:
        description:
            - Determines any rules to be added to a default security group.
            - When creating a network interface, if no security group name is provided, a default security group will be created.
            - If the I(os_type=Windows), a rule allowing RDP access will be added.
            - If the I(os_type=Linux), a rule allowing SSH access will be added.
        type: str
        choices:
            - Windows
            - Linux
        default: Linux
    private_ip_address:
        description:
            - (Deprecate) Valid IPv4 address that falls within the specified subnet.
            - This option will be deprecated in 2.9, use I(ip_configurations) instead.
        type: str
    private_ip_allocation_method:
        description:
            - (Deprecate) Whether or not the assigned IP address is permanent.
            - When creating a network interface, if you specify I(private_ip_address=Static), you must provide a value for I(private_ip_address).
            - You can update the allocation method to C(Static) after a dynamic private IP address has been assigned.
            - This option will be deprecated in 2.9, use I(ip_configurations) instead.
        default: Dynamic
        type: str
        choices:
            - Dynamic
            - Static
    public_ip:
        description:
            - (Deprecate) When creating a network interface, if no public IP address name is provided a default public IP address will be created.
            - Set to C(false) if you do not want a public IP address automatically created.
            - This option will be deprecated in 2.9, use I(ip_configurations) instead.
        type: bool
        default: 'yes'
    public_ip_address_name:
        description:
            - (Deprecate) Name of an existing public IP address object to associate with the security group.
            - This option will be deprecated in 2.9, use I(ip_configurations) instead.
        type: str
        aliases:
            - public_ip_address
            - public_ip_name
    public_ip_allocation_method:
        description:
            - (Deprecate) If a I(public_ip_address_name) is not provided, a default public IP address will be created.
            - The allocation method determines whether or not the public IP address assigned to the network interface is permanent.
            - This option will be deprecated in 2.9, use I(ip_configurations) instead.
        type: str
        choices:
            - Dynamic
            - Static
        default: Dynamic
    ip_configurations:
        description:
            - List of IP configurations. Each configuration object should include
              field I(private_ip_address), I(private_ip_allocation_method), I(public_ip_address_name), I(public_ip), I(public_ip_allocation_method), I(name).
        type: list
        elements: dict
        default: []
        suboptions:
            name:
                description:
                    - Name of the IP configuration.
                required: true
                type: str
            private_ip_address:
                description:
                    - Private IP address for the IP configuration.
                type: str
            private_ip_address_version:
                description:
                    - The version of the IP configuration.
                type: str
                choices:
                    - IPv4
                    - IPv6
                default: IPv4
            private_ip_allocation_method:
                description:
                    - Private IP allocation method.
                type: str
                choices:
                    - Dynamic
                    - Static
                default: Dynamic
            public_ip_address_name:
                description:
                    - Name of the public IP address. None for disable IP address.
                type: str
                aliases:
                    - public_ip_address
                    - public_ip_name
            public_ip_allocation_method:
                description:
                    - Public IP allocation method.
                type: str
                choices:
                    - Dynamic
                    - Static
                default: Dynamic
            load_balancer_backend_address_pools:
                description:
                    - List of existing load-balancer backend address pools to associate with the network interface.
                    - Can be written as a resource ID.
                    - Also can be a dict of I(name) and I(load_balancer).
                type: list
                elements: raw
            application_gateway_backend_address_pools:
                description:
                    - List of existing application gateway backend address pools to associate with the network interface.
                    - Can be written as a resource ID.
                    - Also can be a dict of I(name) and I(application_gateway).
                type: list
                elements: raw
                version_added: "1.10.0"
            primary:
                description:
                    - Whether the IP configuration is the primary one in the list.
                    - The first IP configuration default set to I(primary=True).
                type: bool
                default: False
            application_security_groups:
                description:
                    - List of application security groups in which the IP configuration is included.
                    - Element of the list could be a resource id of application security group, or dict of I(resource_group) and I(name).
                type: list
                elements: raw
    enable_accelerated_networking:
        description:
            - Whether the network interface should be created with the accelerated networking feature or not.
        type: bool
        default: False
    create_with_security_group:
        description:
            - Whether a security group should be be created with the NIC.
            - If this flag set to C(True) and no I(security_group) set, a default security group will be created.
        type: bool
        default: True
    security_group:
        description:
            - An existing security group with which to associate the network interface.
            - If not provided, a default security group will be created when I(create_with_security_group=true).
            - It can be the name of security group.
            - Make sure the security group is in the same resource group when you only give its name.
            - It can be the resource id.
            - It can be a dict contains security_group's I(name) and I(resource_group).
        aliases:
            - security_group_name
        type: raw
    open_ports:
        description:
            - When a default security group is created for a Linux host a rule will be added allowing inbound TCP
              connections to the default SSH port C(22), and for a Windows host rules will be added allowing inbound
              access to RDP ports C(3389) and C(5986). Override the default ports by providing a list of open ports.
        type: list
        elements: str
    enable_ip_forwarding:
        description:
            - Whether to enable IP forwarding.
        aliases:
            - ip_forwarding
        type: bool
        default: False
    dns_servers:
        description:
            - Which DNS servers should the NIC lookup.
            - List of IP addresses.
        type: list
        elements: str
extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Chris Houseknecht (@chouseknecht)
    - Matt Davis (@nitzmahone)
    - Yuwei Zhou (@yuwzho)
'''

EXAMPLES = '''
- name: Create a network interface with minimal parameters
  azure_rm_networkinterface:
    name: nic001
    resource_group: myResourceGroup
    virtual_network: vnet001
    subnet_name: subnet001
    ip_configurations:
      - name: ipconfig1
        public_ip_address_name: publicip001
        primary: true

- name: Create a network interface with private IP address only (no Public IP)
  azure_rm_networkinterface:
    name: nic001
    resource_group: myResourceGroup
    virtual_network: vnet001
    subnet_name: subnet001
    create_with_security_group: false
    ip_configurations:
      - name: ipconfig1
        primary: true

- name: Create a network interface for use in a Windows host (opens RDP port) with custom RDP port
  azure_rm_networkinterface:
    name: nic002
    resource_group: myResourceGroup
    virtual_network: vnet001
    subnet_name: subnet001
    os_type: Windows
    rdp_port: 3399
    security_group: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/myResourceGroup/providers/Microsoft.Network/networkSecurit
                     yGroups/nsg001"
    ip_configurations:
      - name: ipconfig1
        public_ip_address_name: publicip001
        primary: true

- name: Create a network interface using existing security group and public IP
  azure_rm_networkinterface:
    name: nic003
    resource_group: myResourceGroup
    virtual_network: vnet001
    subnet_name: subnet001
    security_group: secgroup001
    ip_configurations:
      - name: ipconfig1
        public_ip_address_name: publicip001
        primary: true

- name: Create a network with multiple ip configurations
  azure_rm_networkinterface:
    name: nic004
    resource_group: myResourceGroup
    subnet_name: subnet001
    virtual_network: vnet001
    security_group:
      name: testnic002
      resource_group: Testing1
    ip_configurations:
      - name: ipconfig1
        public_ip_address_name: publicip001
        primary: true
      - name: ipconfig2
        load_balancer_backend_address_pools:
          - "{{ loadbalancer001.state.backend_address_pools[0].id }}"
          - name: backendaddrpool1
            load_balancer: loadbalancer001

- name: Create network interface attached to application gateway backend address pool
  azure_rm_networkinterface:
    name: nic-appgw
    resource_group: myResourceGroup
    virtual_network: vnet001
    subnet_name: subnet001
    create_with_security_group: false
    public_ip: false
    ip_configurations:
      - name: default
        primary: true
        application_gateway_backend_address_pools:
          - name: myApplicationGatewayBackendAddressPool
            application_gateway: myApplicationGateway

- name: Create a network interface in accelerated networking mode
  azure_rm_networkinterface:
    name: nic005
    resource_group: myResourceGroup
    virtual_network_name: vnet001
    subnet_name: subnet001
    enable_accelerated_networking: true

- name: Create a network interface with IP forwarding
  azure_rm_networkinterface:
    name: nic001
    resource_group: myResourceGroup
    virtual_network: vnet001
    subnet_name: subnet001
    ip_forwarding: true
    ip_configurations:
      - name: ipconfig1
        public_ip_address_name: publicip001
        primary: true

- name: Create a network interface with dns servers
  azure_rm_networkinterface:
    name: nic009
    resource_group: myResourceGroup
    virtual_network: vnet001
    subnet_name: subnet001
    dns_servers:
      - 8.8.8.8

- name: Delete network interface
  azure_rm_networkinterface:
    resource_group: myResourceGroup
    name: nic003
    state: absent
'''

RETURN = '''
state:
    description:
        - The current state of the network interface.
    returned: always
    type: complex
    contains:
        dns_server:
            description:
                - Which DNS servers should the NIC lookup.
                - List of IP addresses.
            type: list
            sample: ['8.9.10.11', '7.8.9.10']
        dns_setting:
            description:
                - The DNS settings in network interface.
            type: dict
            sample: {
                "applied_dns_servers": [],
                "dns_servers": [
                    "8.9.10.11",
                    "7.8.9.10"
                ],
                "internal_dns_name_label": null,
                "internal_fqdn": null
                }
        enable_ip_forwarding:
            description:
                Whether to enable IP forwarding.
            type: bool
            sample: true
        etag:
            description:
                - A unique read-only string that changes whenever the resource is updated.
            type: str
            sample: 'W/"be115a43-2148-4545-a324-f33ad444c926"'
        id:
            description:
                - Id of the network interface.
            type: str
            sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/myResourceGroup/providers/Microsoft.Network/networkInterfaces/nic003"
        enable_accelerated_networking:
            description:
                 - Whether the network interface should be created with the accelerated networking feature or not.
            type: bool
            sample: true
        ip_configurations:
            description:
                - List of IP configurations.
            type: complex
            contains:
                name:
                    description:
                        - Name of the IP configuration.
                    type: str
                    sample: default
                load_balancer_backend_address_pools:
                    description:
                        - List of existing load-balancer backend address pools associated with the network interface.
                    type: list
                application_gateway_backend_address_pools:
                    description:
                        - List of existing application gateway backend address pool resource IDs associated with the network interface.
                    type: list
                    version_added: "1.10.0"
                    sample: ["/subscriptions/xxx/resourceGroups/myResourceGroup/providers/Microsoft.Network/applicationGateways/myGateway/
                        backendAddressPools/myBackendAddressPool"]
                private_ip_address:
                    description:
                        - Private IP address for the IP configuration.
                    type: str
                    sample: "10.1.0.10"
                private_ip_address_version:
                    description:
                        - The version of the IP configuration.
                    type: str
                    sample: "IPv4"
                private_ip_allocation_method:
                    description:
                        - Private IP allocation method.
                    type: str
                    sample: "Static"
                public_ip_address:
                    description:
                        - Name of the public IP address. None for disable IP address.
                    type: dict
                    sample: {
                        "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/myResourceGroup/providers/Microsoft.Network/publicIPAddresse
                        s/publicip001",
                        "name": "publicip001"
                        }
                subnet:
                    description:
                        - The reference of the subnet resource.
                    type: dict
                    sample: {
                        "id": "/subscriptions/xxxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/
                         myresourcegroup/providers/Microsoft.Network/virtualNetworks/tnb57dc95318/subnets/tnb57dc95318",
                        "name": "tnb57dc95318",
                        "resource_group": "myresourcegroup",
                        "virtual_network_name": "tnb57dc95318"
                        }
        location:
            description:
                - The network interface resource location.
            type: str
            sample: eastus
        mac_address:
            description:
                - The MAC address of the network interface.
            type: str
        name:
            description:
                - Name of the network interface.
            type: str
            sample: nic003
        network_security_group:
            description:
                - The reference of the network security group resource.
            type: dict
            sample: {
                "id": "/subscriptions//xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroup/
                myResourceGroup/providers/Microsoft.Network/networkSecurityGroups/nsg001",
                "name": "nsg001"
                }
        primary:
            description:
                - Get whether this is a primary network interface on virtual machine.
            type: bool
            sample: true
        provisioning_state:
            description:
                - The provisioning state of the public IP resource.
            type: str
            sample: Succeeded
        tags:
            description:
                -Tags of the network interface.
            type: dict
            sample: { 'key': 'value' }
        type:
            description:
                - Type of the resource.
            type: str
            sample: "Microsoft.Network/networkInterfaces"
'''

try:
    from azure.mgmt.core.tools import resource_id, is_valid_resource_id
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    # This is handled in azure_rm_common
    pass

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import (AzureRMModuleBase,
                                                                                         azure_id_to_dict,
                                                                                         normalize_location_name,
                                                                                         format_resource_id
                                                                                         )
from ansible.module_utils._text import to_native


def subnet_to_dict(subnet):
    dic = azure_id_to_dict(subnet.id)
    return dict(
        id=subnet.id,
        virtual_network_name=dic.get('virtualNetworks'),
        resource_group=dic.get('resourceGroups'),
        name=dic.get('subnets')
    )


def nic_to_dict(nic):
    ip_configurations = [
        dict(
            name=config.name,
            private_ip_address=config.private_ip_address,
            private_ip_address_version=config.private_ip_address_version,
            private_ip_allocation_method=config.private_ip_allocation_method,
            subnet=subnet_to_dict(config.subnet),
            primary=config.primary if config.primary else False,
            load_balancer_backend_address_pools=([item.id for item in config.load_balancer_backend_address_pools]
                                                 if config.load_balancer_backend_address_pools else None),
            application_gateway_backend_address_pools=([item.id for item in config.application_gateway_backend_address_pools]
                                                       if config.application_gateway_backend_address_pools else None),
            public_ip_address=dict(
                id=config.public_ip_address.id,
                name=azure_id_to_dict(config.public_ip_address.id).get('publicIPAddresses'),
                public_ip_allocation_method=config.public_ip_address.public_ip_allocation_method
            ) if config.public_ip_address else None,
            application_security_groups=([asg.id for asg in config.application_security_groups]
                                         if config.application_security_groups else None)
        ) for config in nic.ip_configurations
    ]
    return dict(
        id=nic.id,
        name=nic.name,
        type=nic.type,
        location=nic.location,
        tags=nic.tags,
        network_security_group=dict(
            id=nic.network_security_group.id,
            name=azure_id_to_dict(nic.network_security_group.id).get('networkSecurityGroups')
        ) if nic.network_security_group else None,
        dns_settings=dict(
            dns_servers=nic.dns_settings.dns_servers,
            applied_dns_servers=nic.dns_settings.applied_dns_servers,
            internal_dns_name_label=nic.dns_settings.internal_dns_name_label,
            internal_fqdn=nic.dns_settings.internal_fqdn
        ),
        ip_configurations=ip_configurations,
        ip_configuration=ip_configurations[0] if len(ip_configurations) == 1 else None,  # for compatible issue, keep this field
        mac_address=nic.mac_address,
        enable_ip_forwarding=nic.enable_ip_forwarding,
        provisioning_state=nic.provisioning_state,
        etag=nic.etag,
        enable_accelerated_networking=nic.enable_accelerated_networking,
        dns_servers=nic.dns_settings.dns_servers,
    )


ip_configuration_spec = dict(
    name=dict(type='str', required=True),
    private_ip_address=dict(type='str'),
    private_ip_address_version=dict(type='str', choices=['IPv4', 'IPv6'], default='IPv4'),
    private_ip_allocation_method=dict(type='str', choices=['Dynamic', 'Static'], default='Dynamic'),
    public_ip_address_name=dict(type='str', aliases=['public_ip_address', 'public_ip_name']),
    public_ip_allocation_method=dict(type='str', choices=['Dynamic', 'Static'], default='Dynamic'),
    load_balancer_backend_address_pools=dict(type='list', elements='raw'),
    application_gateway_backend_address_pools=dict(type='list', elements='raw'),
    primary=dict(type='bool', default=False),
    application_security_groups=dict(type='list', elements='raw')
)


class AzureRMNetworkInterface(AzureRMModuleBase):

    def __init__(self):

        self.module_arg_spec = dict(
            resource_group=dict(type='str', required=True),
            name=dict(type='str', required=True),
            location=dict(type='str'),
            enable_accelerated_networking=dict(type='bool', default=False),
            create_with_security_group=dict(type='bool', default=True),
            security_group=dict(type='raw', aliases=['security_group_name']),
            state=dict(default='present', choices=['present', 'absent']),
            private_ip_address=dict(type='str'),
            private_ip_allocation_method=dict(type='str', choices=['Dynamic', 'Static'], default='Dynamic'),
            public_ip_address_name=dict(type='str', aliases=['public_ip_address', 'public_ip_name']),
            public_ip=dict(type='bool', default=True),
            subnet_name=dict(type='str', aliases=['subnet']),
            virtual_network=dict(type='raw', aliases=['virtual_network_name']),
            public_ip_allocation_method=dict(type='str', choices=['Dynamic', 'Static'], default='Dynamic'),
            ip_configurations=dict(type='list', default=[], elements='dict', options=ip_configuration_spec),
            os_type=dict(type='str', choices=['Windows', 'Linux'], default='Linux'),
            open_ports=dict(type='list', elements='str'),
            enable_ip_forwarding=dict(type='bool', aliases=['ip_forwarding'], default=False),
            dns_servers=dict(type='list', elements='str'),
        )

        required_if = [
            ('state', 'present', ['subnet_name', 'virtual_network'])
        ]

        self.resource_group = None
        self.name = None
        self.location = None
        self.create_with_security_group = None
        self.enable_accelerated_networking = None
        self.security_group = None
        self.private_ip_address = None
        self.private_ip_allocation_method = None
        self.public_ip_address_name = None
        self.public_ip = None
        self.subnet_name = None
        self.virtual_network = None
        self.public_ip_allocation_method = None
        self.state = None
        self.tags = None
        self.os_type = None
        self.open_ports = None
        self.enable_ip_forwarding = None
        self.ip_configurations = None
        self.dns_servers = None

        self.results = dict(
            changed=False,
            state=dict(),
        )

        super(AzureRMNetworkInterface, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                      supports_check_mode=True,
                                                      required_if=required_if)

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec.keys()) + ['tags']:
            setattr(self, key, kwargs[key])

        results = None
        changed = False
        nic = None
        nsg = None

        resource_group = self.get_resource_group(self.resource_group)
        if not self.location:
            # Set default location
            self.location = resource_group.location
        self.location = normalize_location_name(self.location)

        # parse the virtual network resource group and name
        self.virtual_network = self.parse_resource_to_dict(self.virtual_network)

        # if not set the security group name, use nic name for default
        self.security_group = self.parse_resource_to_dict(self.security_group or self.name)

        # if application security groups set, convert to resource id format
        if self.ip_configurations:
            for config in self.ip_configurations:
                if config.get('application_security_groups'):
                    asgs = []
                    for asg in config['application_security_groups']:
                        asg_resource_id = asg
                        if isinstance(asg, str) and (not is_valid_resource_id(asg)):
                            asg = self.parse_resource_to_dict(asg)
                        if isinstance(asg, dict):
                            asg_resource_id = format_resource_id(val=asg['name'],
                                                                 subscription_id=self.subscription_id,
                                                                 namespace='Microsoft.Network',
                                                                 types='applicationSecurityGroups',
                                                                 resource_group=asg['resource_group'])
                        asgs.append(asg_resource_id)
                    if len(asgs) > 0:
                        config['application_security_groups'] = asgs

        if self.state == 'present' and not self.ip_configurations:
            # construct the ip_configurations array for compatible
            self.deprecate('Setting ip_configuration flatten is deprecated and will be removed.'
                           ' Using ip_configurations list to define the ip configuration', version=(2, 9))
            self.ip_configurations = [
                dict(
                    private_ip_address=self.private_ip_address,
                    private_ip_allocation_method=self.private_ip_allocation_method,
                    public_ip_address_name=self.public_ip_address_name if self.public_ip else None,
                    public_ip_allocation_method=self.public_ip_allocation_method,
                    name='default',
                    primary=True
                )
            ]

        try:
            self.log('Fetching network interface {0}'.format(self.name))
            nic = self.network_client.network_interfaces.get(self.resource_group, self.name)

            self.log('Network interface {0} exists'.format(self.name))
            self.check_provisioning_state(nic, self.state)
            results = nic_to_dict(nic)
            self.log(results, pretty_print=True)

            nsg = None
            if self.state == 'present':
                # check for update
                update_tags, results['tags'] = self.update_tags(results['tags'])
                if update_tags:
                    changed = True

                if self.create_with_security_group != bool(results.get('network_security_group')):
                    self.log("CHANGED: add or remove network interface {0} network security group".format(self.name))
                    changed = True

                if self.enable_accelerated_networking != bool(results.get('enable_accelerated_networking')):
                    self.log("CHANGED: Accelerated Networking set to {0} (previously {1})".format(
                        self.enable_accelerated_networking,
                        results.get('enable_accelerated_networking')))
                    changed = True

                if self.enable_ip_forwarding != bool(results.get('enable_ip_forwarding')):
                    self.log("CHANGED: IP forwarding set to {0} (previously {1})".format(
                        self.enable_ip_forwarding,
                        results.get('enable_ip_forwarding')))
                    changed = True

                # We need to ensure that dns_servers are list like
                dns_servers_res = results.get('dns_settings').get('dns_servers')
                _dns_servers_set = sorted(self.dns_servers) if isinstance(self.dns_servers, list) else list()
                _dns_servers_res = sorted(dns_servers_res) if isinstance(self.dns_servers, list) else list()
                if _dns_servers_set != _dns_servers_res:
                    self.log("CHANGED: DNS servers set to {0} (previously {1})".format(
                        ", ".join(_dns_servers_set),
                        ", ".join(_dns_servers_res)))
                    changed = True

                if not changed:
                    nsg = self.get_security_group(self.security_group['resource_group'], self.security_group['name'])
                    if nsg and results.get('network_security_group') and results['network_security_group'].get('id') != nsg.id:
                        self.log("CHANGED: network interface {0} network security group".format(self.name))
                        changed = True

                if results['ip_configurations'][0]['subnet']['virtual_network_name'] != self.virtual_network['name']:
                    self.log("CHANGED: network interface {0} virtual network name".format(self.name))
                    changed = True

                if results['ip_configurations'][0]['subnet']['resource_group'] != self.virtual_network['resource_group']:
                    self.log("CHANGED: network interface {0} virtual network resource group".format(self.name))
                    changed = True

                if results['ip_configurations'][0]['subnet']['name'] != self.subnet_name:
                    self.log("CHANGED: network interface {0} subnet name".format(self.name))
                    changed = True

                # check the ip_configuration is changed
                # construct two set with the same structure and then compare
                # the list should contains:
                # name, private_ip_address, public_ip_address_name, private_ip_allocation_method, subnet_name
                ip_configuration_result = self.construct_ip_configuration_set(results['ip_configurations'])
                ip_configuration_request = self.construct_ip_configuration_set(self.ip_configurations)
                ip_configuration_result_name = [item['name'] for item in ip_configuration_result]
                for item_request in ip_configuration_request:
                    if item_request['name'] not in ip_configuration_result_name:
                        changed = True
                        break
                    else:
                        for item_result in ip_configuration_result:
                            if len(ip_configuration_request) == 1 and len(ip_configuration_result) == 1:
                                item_request['primary'] = True
                            if item_request['name'] == item_result['name'] and item_request != item_result:
                                changed = True
                                break

            elif self.state == 'absent':
                self.log("CHANGED: network interface {0} exists but requested state is 'absent'".format(self.name))
                changed = True
        except ResourceNotFoundError:
            self.log('Network interface {0} does not exist'.format(self.name))
            if self.state == 'present':
                self.log("CHANGED: network interface {0} does not exist but requested state is 'present'".format(self.name))
                changed = True

        self.results['changed'] = changed
        self.results['state'] = results

        if self.check_mode:
            return self.results

        if changed:
            if self.state == 'present':
                subnet = self.network_models.SubResource(
                    id='/subscriptions/{0}/resourceGroups/{1}/providers/Microsoft.Network/virtualNetworks/{2}/subnets/{3}'.format(
                        self.virtual_network['subscription_id'],
                        self.virtual_network['resource_group'],
                        self.virtual_network['name'],
                        self.subnet_name))

                nic_ip_configurations = [
                    self.network_models.NetworkInterfaceIPConfiguration(
                        private_ip_allocation_method=ip_config.get('private_ip_allocation_method'),
                        private_ip_address=ip_config.get('private_ip_address'),
                        private_ip_address_version=ip_config.get('private_ip_address_version'),
                        name=ip_config.get('name'),
                        subnet=subnet,
                        public_ip_address=self.get_or_create_public_ip_address(ip_config),
                        load_balancer_backend_address_pools=([self.network_models.BackendAddressPool(id=self.backend_addr_pool_id(bap_id))
                                                              for bap_id in ip_config.get('load_balancer_backend_address_pools')]
                                                             if ip_config.get('load_balancer_backend_address_pools') else None),
                        application_gateway_backend_address_pools=([self.network_models.ApplicationGatewayBackendAddressPool
                                                                    (id=self.gateway_backend_addr_pool_id(bap_id))
                                                                    for bap_id in ip_config.get('application_gateway_backend_address_pools')]
                                                                   if ip_config.get('application_gateway_backend_address_pools') else None),
                        primary=ip_config.get('primary'),
                        application_security_groups=([self.network_models.ApplicationSecurityGroup(id=asg_id)
                                                      for asg_id in ip_config.get('application_security_groups')]
                                                     if ip_config.get('application_security_groups') else None)
                    ) for ip_config in self.ip_configurations
                ]

                nsg = self.create_default_securitygroup(self.security_group['resource_group'],
                                                        self.location,
                                                        self.security_group['name'],
                                                        self.os_type,
                                                        self.open_ports) if self.create_with_security_group else None

                self.log('Creating or updating network interface {0}'.format(self.name))
                nic = self.network_models.NetworkInterface(
                    id=results['id'] if results else None,
                    location=self.location,
                    tags=self.tags,
                    ip_configurations=nic_ip_configurations,
                    enable_accelerated_networking=self.enable_accelerated_networking,
                    enable_ip_forwarding=self.enable_ip_forwarding,
                    network_security_group=nsg
                )
                if self.dns_servers:
                    dns_settings = self.network_models.NetworkInterfaceDnsSettings(
                        dns_servers=self.dns_servers)
                    nic.dns_settings = dns_settings
                self.results['state'] = self.create_or_update_nic(nic)
            elif self.state == 'absent':
                self.log('Deleting network interface {0}'.format(self.name))
                self.delete_nic()
                # Delete doesn't return anything. If we get this far, assume success
                self.results['state']['status'] = 'Deleted'

        return self.results

    def get_or_create_public_ip_address(self, ip_config):
        name = ip_config.get('public_ip_address_name')

        if not (self.public_ip and name):
            return None

        pip = self.get_public_ip_address(name)
        if not pip:
            params = self.network_models.PublicIPAddress(
                location=self.location,
                public_ip_allocation_method=ip_config.get('public_ip_allocation_method'),
            )
            try:
                poller = self.network_client.public_ip_addresses.begin_create_or_update(self.resource_group, name, params)
                pip = self.get_poller_result(poller)
            except Exception as exc:
                self.fail("Error creating {0} - {1}".format(name, str(exc)))
        return pip

    def create_or_update_nic(self, nic):
        try:
            poller = self.network_client.network_interfaces.begin_create_or_update(self.resource_group, self.name, nic)
            new_nic = self.get_poller_result(poller)
            return nic_to_dict(new_nic)
        except Exception as exc:
            self.fail("Error creating or updating network interface {0} - {1}".format(self.name, str(exc)))

    def delete_nic(self):
        try:
            poller = self.network_client.network_interfaces.begin_delete(self.resource_group, self.name)
            self.get_poller_result(poller)
        except Exception as exc:
            self.fail("Error deleting network interface {0} - {1}".format(self.name, str(exc)))
        return True

    def get_public_ip_address(self, name):
        self.log("Fetching public ip address {0}".format(name))
        try:
            return self.network_client.public_ip_addresses.get(self.resource_group, name)
        except ResourceNotFoundError as exc:
            return None

    def get_security_group(self, resource_group, name):
        self.log("Fetching security group {0}".format(name))
        try:
            return self.network_client.network_security_groups.get(resource_group, name)
        except ResourceNotFoundError as exc:
            return None

    def backend_addr_pool_id(self, val):
        if isinstance(val, dict):
            lb = val.get('load_balancer', None)
            name = val.get('name', None)
            if lb and name:
                return resource_id(subscription=self.subscription_id,
                                   resource_group=self.resource_group,
                                   namespace='Microsoft.Network',
                                   type='loadBalancers',
                                   name=lb,
                                   child_type_1='backendAddressPools',
                                   child_name_1=name)
        return val

    def gateway_backend_addr_pool_id(self, val):
        if isinstance(val, dict):
            appgw = val.get('application_gateway', None)
            name = val.get('name', None)
            if appgw and name:
                return resource_id(subscription=self.subscription_id,
                                   resource_group=self.resource_group,
                                   namespace='Microsoft.Network',
                                   type='applicationGateways',
                                   name=appgw,
                                   child_type_1='backendAddressPools',
                                   child_name_1=name)
        return val

    def construct_ip_configuration_set(self, raw):
        configurations = [dict(
            private_ip_allocation_method=to_native(item.get('private_ip_allocation_method')),
            public_ip_address_name=(to_native(item.get('public_ip_address').get('name'))
                                    if item.get('public_ip_address') else to_native(item.get('public_ip_address_name'))),
            primary=item.get('primary'),
            load_balancer_backend_address_pools=(set([to_native(self.backend_addr_pool_id(id))
                                                      for id in item.get('load_balancer_backend_address_pools')])
                                                 if item.get('load_balancer_backend_address_pools') else None),
            application_gateway_backend_address_pools=(set([to_native(self.gateway_backend_addr_pool_id(id))
                                                           for id in item.get('application_gateway_backend_address_pools')])
                                                       if item.get('application_gateway_backend_address_pools') else None),
            application_security_groups=(set([to_native(asg_id) for asg_id in item.get('application_security_groups')])
                                         if item.get('application_security_groups') else None),
            name=to_native(item.get('name'))
        ) for item in raw]
        return configurations


def main():
    AzureRMNetworkInterface()


if __name__ == '__main__':
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The ios vlans fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type


import re

from copy import deepcopy

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.argspec.vlans.vlans import (
    VlansArgs,
)


class VlansFacts(object):
    """The ios vlans fact class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = VlansArgs.argument_spec
        spec = deepcopy(self.argument_spec)
        if subspec:
            if options:
                facts_argument_spec = spec[subspec][options]
            else:
                facts_argument_spec = spec[subspec]
        else:
            facts_argument_spec = spec

        self.generated_spec = utils.generate_dict(facts_argument_spec)

    def get_vlans_data(self, connection, configuration):
        """Checks device is L2/L3 and returns
        facts gracefully. Does not fail module.
        """
        if configuration:
            cmd = "show running-config | sec ^vlan configuration .+"
        else:
            cmd = "show vlan"
        check_os_type = connection.get_device_info()
        if check_os_type.get("network_os_type") == "L3":
            return ""
        return connection.get(cmd)

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for vlans
        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected conf
        :rtype: dictionary
        :returns: facts
        """
        configuration = self._module.params["configuration"]
        objs = []

        if not data:
            data = self.get_vlans_data(connection, configuration)
        if not configuration:
            objs = self.parse_vlan(data)
        else:
            objs = self.parse_vlan_config(data)

        facts = {}
        if objs:
            facts["vlans"] = []
            params = utils.validate_config(self.argument_spec, {"config": objs})

            for cfg in params["config"]:
                facts["vlans"].append(utils.remove_empties(cfg))
        ansible_facts["ansible_network_resources"].update(facts)
        return ansible_facts

    def render_config(self, spec, conf, vlan_info):
        """
        Render config as dictionary structure and delete keys
          from spec for null values

        :param spec: The facts tree, generated from the argspec
        :param conf: The configuration
        :rtype: dictionary
        :returns: The generated config
        """
        config = deepcopy(spec)

        if vlan_info == "Name" and "VLAN Name" not in conf:
            conf = list(filter(None, conf.split(" ")))
            config["vlan_id"] = int(conf[0])
            config["name"] = conf[1]
            state_idx = 2
            for i in range(2, len(conf)):  # check for index where state starts
                if conf[i] in ["suspended", "active"]:
                    state_idx = i
                    break
                elif conf[i].split("/")[0] in ["sus", "act"]:
                    state_idx = i
                    break
                config["name"] += " " + conf[i]
            try:
                if len(conf[state_idx].split("/")) > 1:
                    _state = conf[state_idx].split("/")[0]
                    if _state == "sus":
                        config["state"] = "suspend"
                    elif _state == "act":
                        config["state"] = "active"
                    config["shutdown"] = "enabled"
                else:
                    if conf[state_idx] == "suspended":
                        config["state"] = "suspend"
                    elif conf[state_idx] == "active":
                        config["state"] = "active"
                    config["shutdown"] = "disabled"
            except IndexError:
                pass
        elif vlan_info == "Type" and "VLAN Type" not in conf:
            conf = list(filter(None, conf.split(" ")))
            config["mtu"] = int(conf[3])
        elif vlan_info == "Remote":
            if len(conf.split(",")) > 1 or conf.isdigit():
                remote_span_vlan = []
                if len(conf.split(",")) > 1:
                    remote_span_vlan = conf.split(",")
                else:
                    remote_span_vlan.append(conf)
                remote_span = []
                for each in remote_span_vlan:
                    split_sp_list = each.split("-")
                    if len(split_sp_list) > 1:  # break range
                        for r_sp in range(int(split_sp_list[0]), int(split_sp_list[1]) + 1):
                            remote_span.append(r_sp)
                    else:
                        remote_span.append(int(each))
                config["remote_span"] = remote_span

        elif vlan_info == "Private" and "Primary Secondary" not in conf:
            conf = list(filter(None, conf.split(" ")))

            pri_idx = 0
            sec_idx = 1
            priv_type_idx = 2

            config["tmp_pvlans"] = {
                "primary": conf[pri_idx],
                "secondary": conf[sec_idx],
                "sec_type": conf[priv_type_idx],
            }
        return utils.remove_empties(config)

    def parse_vlan_config(self, vlan_conf):
        vlan_list = list()

        re1 = re.compile(r"^vlan configuration +(?P<vlan>\d+)$")
        re2 = re.compile(r"^member +(evpn\-instance +(?P<evi>\d+) )?vni (?P<vni>[\d\-]+)$")

        for line in vlan_conf.splitlines():
            line = line.strip()

            m = re1.match(line)
            if m:
                vlan = m.groupdict()["vlan"]
                vlan_dict = {"vlan_id": vlan}
                continue

            m = re2.match(line)
            if m:
                group = m.groupdict()
                vlan_dict.update({"member": {}})
                vlan_dict["member"].update({"vni": group["vni"]})
                if group["evi"]:
                    vlan_dict["member"].update({"evi": group["evi"]})
                vlan_list.append(vlan_dict)

        return vlan_list

    def parse_vlan(self, data):
        objs = []
        mtu_objs = []
        remote_objs = []
        final_objs = []
        pvlan_objs = []

        # operate on a collection of resource x
        config = data.split("\n")
        # Get individual vlan configs separately
        vlan_info = ""
        temp = ""
        vlan_name = True
        for conf in config:
            if len(list(filter(None, conf.split(" ")))) <= 2 and vlan_name:
                temp = temp + conf
                if len(list(filter(None, temp.split(" ")))) <= 2:
                    continue
            if "VLAN Name" in conf:
                vlan_info = "Name"
            elif "VLAN Type" in conf:
                vlan_info = "Type"
                vlan_name = False
            elif "Remote SPAN" in conf:
                vlan_info = "Remote"
                vlan_name = False
            elif "VLAN AREHops" in conf or "STEHops" in conf:
                vlan_info = "Hops"
                vlan_name = False
            elif "Primary Secondary" in conf:
                vlan_info = "Private"
                vlan_name = False
            if temp:
                conf = temp
                temp = ""
            if conf and " " not in filter(None, conf.split("-")) and not conf.split(" ")[0] == "":
                obj = self.render_config(self.generated_spec, conf, vlan_info)
                if "mtu" in obj:
                    mtu_objs.append(obj)
                elif "remote_span" in obj:
                    remote_objs = obj
                elif "tmp_pvlans" in obj:
                    pvlan_objs.append(obj)
                elif obj:
                    objs.append(obj)
        # Appending MTU value to the retrieved dictionary
        for o, m in zip(objs, mtu_objs):
            o.update(m)
            final_objs.append(o)

        # Appending Remote Span value to related VLAN
        if remote_objs:
            if remote_objs.get("remote_span"):
                for each in remote_objs.get("remote_span"):
                    for every in final_objs:
                        if each == every.get("vlan_id"):
                            every.update({"remote_span": True})
                            break

        # Appending private vlan information to related VLAN
        if pvlan_objs:
            pvlan_final = {}
            if len(pvlan_objs) > 0:
                # Sanitize and structure everything

                for data in pvlan_objs:
                    pvdata = data.get("tmp_pvlans")
                    privlan = pvdata.get("primary")
                    secvlan = pvdata.get("secondary")
                    sectype = pvdata.get("sec_type")

                    # Define the secondary VLAN's type
                    if secvlan and (isinstance(secvlan, int) or secvlan.isnumeric()):
                        secvlan = int(secvlan)
                        pvlan_final[secvlan] = {"private_vlan": {"type": sectype}}

                    # Assemble and merge data for primary private VLANs
                    if privlan and (isinstance(privlan, int) or privlan.isnumeric()):
                        privlan = int(privlan)
                        if privlan not in pvlan_final.keys():
                            pvlan_final[privlan] = {
                                "private_vlan": {"type": "primary", "associated": []},
                            }
                        if secvlan and (isinstance(secvlan, int) or secvlan.isnumeric()):
                            pvlan_final[privlan]["private_vlan"]["associated"].append(int(secvlan))

                # Associate with the proper VLAN in final_objs
                for vlan_id, data in pvlan_final.items():
                    for every in final_objs:
                        if vlan_id == every.get("vlan_id"):
                            every.update(data)

        if final_objs:
            return objs
        else:
            return {}



#
# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

"""
The pop_ace plugin code
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type

from ansible.errors import AnsibleFilterError


def _raise_error(msg):
    """Raise an error message, prepend with filter name
    :param msg: The message
    :type msg: str
    :raises: AnsibleError
    """
    error = "Error when using plugin 'pop_ace': {msg}".format(msg=msg)
    raise AnsibleFilterError(error)


def fail_missing(racl, fail):
    if fail and racl == []:
        _raise_error("no entries removed on the provided match_criteria")


def check_match(ace, match_criteria, match_all, name, afi):
    check_arr = []
    for k, v in match_criteria.items():
        if v:
            if k not in ["source", "destination", "acl_name", "afi"]:
                check_arr.append(True) if ace.get(k, "NA") == match_criteria.get(
                    k,
                ) else check_arr.append(False)
            elif k == "acl_name":
                check_arr.append(True) if name == match_criteria.get(
                    k,
                ) else check_arr.append(False)
            elif k == "afi":
                check_arr.append(True) if afi == match_criteria.get(
                    k,
                ) else check_arr.append(False)
            else:  # for source and destination address
                _sub = "source" if "source" in k else "destination"
                _valid = []
                _valid.append(True) if ace.get(_sub, {}).get("address", "NA") == match_criteria.get(
                    k,
                ) else _valid.append(False)
                _valid.append(True) if ace.get(_sub, {}).get("host", "NA") == match_criteria.get(
                    k,
                ) else _valid.append(False)
                _valid.append(True) if ace.get(_sub, {}).get("any", "NA") == (
                    match_criteria.get(k) == "any"
                ) else _valid.append(False)
                check_arr.append(any(_valid))

    if match_all:  # forces all criteria to match
        return all(check_arr)
    else:
        return any(check_arr)


def _pop_ace(raw_acl, filter_options, match_criteria):
    acls_v4, acls_v6 = [], []
    racls_v4, racls_v6 = [], []

    remove_first_ace_only = True if filter_options.get("remove") == "first" else False
    fail_if_no_match = True if filter_options.get("failed_when") == "missing" else False
    match_all = True if filter_options.get("match_all") is True else False

    final_acl = {
        "acls": [
            {"acls": acls_v4, "afi": "ipv4"},
            {"acls": acls_v6, "afi": "ipv6"},
        ],
    }  # holds final acl data after removal of aces
    rfinal_acl = {
        "acls": [
            {"acls": racls_v4, "afi": "ipv4"},
            {"acls": racls_v6, "afi": "ipv6"},
        ],
    }  # holds removed acl information

    for acls in raw_acl:  # ["acls"]
        afi = acls.get("afi")  # ipv4 or v6

        for acl in acls.get("acls"):
            _aces, _acl = [], {}
            _races, _racl, _rstop = [], {}, True

            aces = acl.get("aces", {})
            name = acl.get("name", "")  # filter by acl_name ignores whole acl entries i.e all aces

            for ace in aces:  # iterate on ace entries
                judge = check_match(
                    ace=ace,
                    match_criteria=match_criteria,
                    match_all=match_all,
                    name=name,
                    afi=afi,
                )
                if judge:  # check matching criteria and remove from final dict
                    if remove_first_ace_only and _rstop:  # removes one ace entry per acl
                        _races.append(ace)
                        _rstop = False
                        continue
                    elif not remove_first_ace_only:  # for remove all
                        _races.append(ace)
                        continue

                _aces.append(
                    ace,
                )

            if _aces:  # store filtered aces
                _acl["name"], _acl["aces"] = name, _aces
                acls_v4.append(_acl) if afi == "ipv4" else acls_v6.append(_acl)

            if _races:  # store removed aces
                _racl["name"], _racl["aces"] = name, _races
                racls_v4.append(_racl) if afi == "ipv4" else racls_v6.append(_racl)

    fail_missing(racls_v4 + racls_v6, fail_if_no_match)

    return final_acl, rfinal_acl


def pop_ace(data, filter_options, match_criteria):
    if not isinstance(data, (list, dict)):
        _raise_error("Input is not valid for pop_ace")
    cleared_data, removed_data = _pop_ace(data, filter_options, match_criteria)
    data = {
        "clean_acls": cleared_data,
        "removed_aces": removed_data,
    }
    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type
from ansible.plugins.action import ActionBase
try:
    from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
        AnsibleArgSpecValidator,
    )
except ImportError:
    ANSIBLE_UTILS_IS_INSTALLED = False
else:
    ANSIBLE_UTILS_IS_INSTALLED = True
from ansible.errors import AnsibleActionFail
from ansible_collections.cisco.dnac.plugins.plugin_utils.dnac import (
    DNACSDK,
    dnac_argument_spec,
    dnac_compare_equality,
    get_dict_result,
)
from ansible_collections.cisco.dnac.plugins.plugin_utils.exceptions import (
    InconsistentParameters,
)

# Get common arguments specification
argument_spec = dnac_argument_spec()
# Add arguments specific for this module
argument_spec.update(dict(
    state=dict(type="str", default="present", choices=["present", "absent"]),
    cdpLevel=dict(type="int"),
    discoveryType=dict(type="str"),
    enablePasswordList=dict(type="list", no_log=True),
    globalCredentialIdList=dict(type="list"),
    httpReadCredential=dict(type="dict"),
    httpWriteCredential=dict(type="dict"),
    ipAddressList=dict(type="str"),
    ipFilterList=dict(type="list"),
    lldpLevel=dict(type="int"),
    name=dict(type="str"),
    netconfPort=dict(type="str"),
    passwordList=dict(type="list", no_log=True),
    preferredMgmtIPMethod=dict(type="str"),
    protocolOrder=dict(type="str"),
    retry=dict(type="int"),
    snmpAuthPassphrase=dict(type="str"),
    snmpAuthProtocol=dict(type="str"),
    snmpMode=dict(type="str"),
    snmpPrivPassphrase=dict(type="str"),
    snmpPrivProtocol=dict(type="str"),
    snmpROCommunity=dict(type="str"),
    snmpROCommunityDesc=dict(type="str"),
    snmpRWCommunity=dict(type="str"),
    snmpRWCommunityDesc=dict(type="str"),
    snmpUserName=dict(type="str"),
    snmpVersion=dict(type="str"),
    timeout=dict(type="int"),
    userNameList=dict(type="list"),
    id=dict(type="str"),
    attributeInfo=dict(type="dict"),
    deviceIds=dict(type="str"),
    discoveryCondition=dict(type="str"),
    discoveryStatus=dict(type="str"),
    isAutoCdp=dict(type="bool"),
    numDevices=dict(type="int"),
    parentDiscoveryId=dict(type="str"),
    retryCount=dict(type="int"),
    snmpRoCommunity=dict(type="str"),
    snmpRoCommunityDesc=dict(type="str"),
    snmpRwCommunity=dict(type="str"),
    snmpRwCommunityDesc=dict(type="str"),
    timeOut=dict(type="int"),
    updateMgmtIp=dict(type="bool"),
))

required_if = [
    ("state", "present", ["id", "name"], True),
    ("state", "absent", ["id", "name"], True),
]
required_one_of = []
mutually_exclusive = []
required_together = []


class Discovery(object):
    def __init__(self, params, dnac):
        self.dnac = dnac
        self.new_object = dict(
            cdpLevel=params.get("cdpLevel"),
            discoveryType=params.get("discoveryType"),
            enablePasswordList=params.get("enablePasswordList"),
            globalCredentialIdList=params.get("globalCredentialIdList"),
            httpReadCredential=params.get("httpReadCredential"),
            httpWriteCredential=params.get("httpWriteCredential"),
            ipAddressList=params.get("ipAddressList"),
            ipFilterList=params.get("ipFilterList"),
            lldpLevel=params.get("lldpLevel"),
            name=params.get("name"),
            netconfPort=params.get("netconfPort"),
            passwordList=params.get("passwordList"),
            preferredMgmtIPMethod=params.get("preferredMgmtIPMethod"),
            protocolOrder=params.get("protocolOrder"),
            retry=params.get("retry"),
            snmpAuthPassphrase=params.get("snmpAuthPassphrase"),
            snmpAuthProtocol=params.get("snmpAuthProtocol"),
            snmpMode=params.get("snmpMode"),
            snmpPrivPassphrase=params.get("snmpPrivPassphrase"),
            snmpPrivProtocol=params.get("snmpPrivProtocol"),
            snmpROCommunity=params.get("snmpROCommunity"),
            snmpROCommunityDesc=params.get("snmpROCommunityDesc"),
            snmpRWCommunity=params.get("snmpRWCommunity"),
            snmpRWCommunityDesc=params.get("snmpRWCommunityDesc"),
            snmpUserName=params.get("snmpUserName"),
            snmpVersion=params.get("snmpVersion"),
            timeout=params.get("timeout"),
            userNameList=params.get("userNameList"),
            id=params.get("id"),
            attributeInfo=params.get("attributeInfo"),
            deviceIds=params.get("deviceIds"),
            discoveryCondition=params.get("discoveryCondition"),
            discoveryStatus=params.get("discoveryStatus"),
            isAutoCdp=params.get("isAutoCdp"),
            numDevices=params.get("numDevices"),
            parentDiscoveryId=params.get("parentDiscoveryId"),
            retryCount=params.get("retryCount"),
            snmpRoCommunity=params.get("snmpRoCommunity"),
            snmpRoCommunityDesc=params.get("snmpRoCommunityDesc"),
            snmpRwCommunity=params.get("snmpRwCommunity"),
            snmpRwCommunityDesc=params.get("snmpRwCommunityDesc"),
            timeOut=params.get("timeOut"),
            updateMgmtIp=params.get("updateMgmtIp"),
        )

    def create_params(self):
        new_object_params = {}
        new_object_params['cdpLevel'] = self.new_object.get('cdpLevel')
        new_object_params['discoveryType'] = self.new_object.get('discoveryType')
        new_object_params['enablePasswordList'] = self.new_object.get('enablePasswordList')
        new_object_params['globalCredentialIdList'] = self.new_object.get('globalCredentialIdList')
        new_object_params['httpReadCredential'] = self.new_object.get('httpReadCredential')
        new_object_params['httpWriteCredential'] = self.new_object.get('httpWriteCredential')
        new_object_params['ipAddressList'] = self.new_object.get('ipAddressList')
        new_object_params['ipFilterList'] = self.new_object.get('ipFilterList')
        new_object_params['lldpLevel'] = self.new_object.get('lldpLevel')
        new_object_params['name'] = self.new_object.get('name')
        new_object_params['netconfPort'] = self.new_object.get('netconfPort')
        new_object_params['passwordList'] = self.new_object.get('passwordList')
        new_object_params['preferredMgmtIPMethod'] = self.new_object.get('preferredMgmtIPMethod')
        new_object_params['protocolOrder'] = self.new_object.get('protocolOrder')
        new_object_params['retry'] = self.new_object.get('retry')
        new_object_params['snmpAuthPassphrase'] = self.new_object.get('snmpAuthPassphrase')
        new_object_params['snmpAuthProtocol'] = self.new_object.get('snmpAuthProtocol')
        new_object_params['snmpMode'] = self.new_object.get('snmpMode')
        new_object_params['snmpPrivPassphrase'] = self.new_object.get('snmpPrivPassphrase')
        new_object_params['snmpPrivProtocol'] = self.new_object.get('snmpPrivProtocol')
        new_object_params['snmpROCommunity'] = self.new_object.get('snmpROCommunity')
        new_object_params['snmpROCommunityDesc'] = self.new_object.get('snmpROCommunityDesc')
        new_object_params['snmpRWCommunity'] = self.new_object.get('snmpRWCommunity')
        new_object_params['snmpRWCommunityDesc'] = self.new_object.get('snmpRWCommunityDesc')
        new_object_params['snmpUserName'] = self.new_object.get('snmpUserName')
        new_object_params['snmpVersion'] = self.new_object.get('snmpVersion')
        new_object_params['timeout'] = self.new_object.get('timeout')
        new_object_params['userNameList'] = self.new_object.get('userNameList')
        return new_object_params

    def delete_by_id_params(self):
        new_object_params = {}
        new_object_params['id'] = self.new_object.get('id')
        return new_object_params

    def convert_list_string(self, pList):
        if isinstance(pList, list):
            if len(pList) > 0:
                pList_str = list(map(str, pList))
                return ', '.join(pList_str)
            else:
                return ''
        else:
            return pList

    def update_all_params(self):
        new_object_params = {}
        new_object_params['attributeInfo'] = self.new_object.get('attributeInfo')
        new_object_params['cdpLevel'] = self.new_object.get('cdpLevel')
        new_object_params['deviceIds'] = self.new_object.get('deviceIds')
        new_object_params['discoveryCondition'] = self.new_object.get('discoveryCondition')
        new_object_params['discoveryStatus'] = self.new_object.get('discoveryStatus')
        new_object_params['discoveryType'] = self.new_object.get('discoveryType')
        new_object_params['enablePasswordList'] = self.convert_list_string(
            self.new_object.get('enablePasswordList')
        )
        new_object_params['globalCredentialIdList'] = self.new_object.get('globalCredentialIdList')
        new_object_params['httpReadCredential'] = self.new_object.get('httpReadCredential')
        new_object_params['httpWriteCredential'] = self.new_object.get('httpWriteCredential')
        new_object_params['id'] = self.new_object.get('id')
        new_object_params['ipAddressList'] = self.new_object.get('ipAddressList')
        new_object_params['ipFilterList'] = self.convert_list_string(
            self.new_object.get('ipFilterList')
        )
        new_object_params['isAutoCdp'] = self.new_object.get('isAutoCdp')
        new_object_params['lldpLevel'] = self.new_object.get('lldpLevel')
        new_object_params['name'] = self.new_object.get('name')
        new_object_params['netconfPort'] = self.new_object.get('netconfPort')
        new_object_params['numDevices'] = self.new_object.get('numDevices')
        new_object_params['parentDiscoveryId'] = self.new_object.get('parentDiscoveryId')
        new_object_params['passwordList'] = self.convert_list_string(
            self.new_object.get('passwordList')
        )
        new_object_params['preferredMgmtIPMethod'] = self.new_object.get('preferredMgmtIPMethod')
        new_object_params['protocolOrder'] = self.new_object.get('protocolOrder')
        new_object_params['retryCount'] = self.new_object.get('retryCount')
        new_object_params['snmpAuthPassphrase'] = self.new_object.get('snmpAuthPassphrase')
        new_object_params['snmpAuthProtocol'] = self.new_object.get('snmpAuthProtocol')
        new_object_params['snmpMode'] = self.new_object.get('snmpMode')
        new_object_params['snmpPrivPassphrase'] = self.new_object.get('snmpPrivPassphrase')
        new_object_params['snmpPrivProtocol'] = self.new_object.get('snmpPrivProtocol')
        new_object_params['snmpRoCommunity'] = self.new_object.get('snmpRoCommunity')
        new_object_params['snmpRoCommunityDesc'] = self.new_object.get('snmpRoCommunityDesc')
        new_object_params['snmpRwCommunity'] = self.new_object.get('snmpRwCommunity')
        new_object_params['snmpRwCommunityDesc'] = self.new_object.get('snmpRwCommunityDesc')
        new_object_params['snmpUserName'] = self.new_object.get('snmpUserName')
        new_object_params['timeOut'] = self.new_object.get('timeOut')
        new_object_params['updateMgmtIp'] = self.new_object.get('updateMgmtIp')
        new_object_params['userNameList'] = self.convert_list_string(
            self.new_object.get('userNameList')
        )
        return new_object_params

    def get_object_by_name(self, name):
        result = None
        # NOTE: Does not have a get by name method or it is in another action
        # NOTE: Does not have get all
        return result

    def get_object_by_id(self, id):
        result = None
        try:
            items = self.dnac.exec(
                family="discovery",
                function="get_discovery_by_id",
                params={"id": id}
            )
            if isinstance(items, dict):
                if 'response' in items:
                    items = items.get('response')
            result = get_dict_result(items, 'id', id)
        except Exception:
            result = None
        return result

    def exists(self):
        id_exists = False
        name_exists = False
        prev_obj = None
        o_id = self.new_object.get("id")
        name = self.new_object.get("name")
        if o_id:
            prev_obj = self.get_object_by_id(o_id)
            id_exists = prev_obj is not None and isinstance(prev_obj, dict)
        if not id_exists and name:
            prev_obj = self.get_object_by_name(name)
            name_exists = prev_obj is not None and isinstance(prev_obj, dict)
        if name_exists:
            _id = prev_obj.get("id")
            if id_exists and name_exists and o_id != _id:
                raise InconsistentParameters("The 'id' and 'name' params don't refer to the same object")
            if _id:
                self.new_object.update(dict(id=_id))
            if _id:
                prev_obj = self.get_object_by_id(_id)
        it_exists = prev_obj is not None and isinstance(prev_obj, dict)
        return (it_exists, prev_obj)

    def requires_update(self, current_obj):
        requested_obj = dict(self.new_object)

        for key in requested_obj.keys():
            if key in ['ipFilterList', 'userNameList']:
                requested_obj[key] = self.convert_list_string(requested_obj.get(key))

        obj_params = [
            ("cdpLevel", "cdpLevel"),
            ("discoveryType", "discoveryType"),
            ("globalCredentialIdList", "globalCredentialIdList"),
            ("httpReadCredential", "httpReadCredential"),
            ("httpWriteCredential", "httpWriteCredential"),
            ("ipAddressList", "ipAddressList"),
            ("ipFilterList", "ipFilterList"),
            ("lldpLevel", "lldpLevel"),
            ("name", "name"),
            ("netconfPort", "netconfPort"),
            ("preferredMgmtIPMethod", "preferredMgmtIPMethod"),
            ("protocolOrder", "protocolOrder"),
            ("retry", "retry"),
            ("snmpAuthPassphrase", "snmpAuthPassphrase"),
            ("snmpAuthProtocol", "snmpAuthProtocol"),
            ("snmpMode", "snmpMode"),
            ("snmpPrivPassphrase", "snmpPrivPassphrase"),
            ("snmpPrivProtocol", "snmpPrivProtocol"),
            ("snmpROCommunity", "snmpROCommunity"),
            ("snmpROCommunityDesc", "snmpROCommunityDesc"),
            ("snmpRWCommunity", "snmpRWCommunity"),
            ("snmpRWCommunityDesc", "snmpRWCommunityDesc"),
            ("snmpUserName", "snmpUserName"),
            ("snmpVersion", "snmpVersion"),
            ("timeout", "timeout"),
            ("userNameList", "userNameList"),
            ("id", "id"),
            ("attributeInfo", "attributeInfo"),
            ("deviceIds", "deviceIds"),
            ("discoveryCondition", "discoveryCondition"),
            ("discoveryStatus", "discoveryStatus"),
            ("isAutoCdp", "isAutoCdp"),
            ("numDevices", "numDevices"),
            ("parentDiscoveryId", "parentDiscoveryId"),
            ("retryCount", "retryCount"),
            ("snmpRoCommunity", "snmpRoCommunity"),
            ("snmpRoCommunityDesc", "snmpRoCommunityDesc"),
            ("snmpRwCommunity", "snmpRwCommunity"),
            ("snmpRwCommunityDesc", "snmpRwCommunityDesc"),
            ("timeOut", "timeOut"),
            ("updateMgmtIp", "updateMgmtIp"),
        ]
        # Method 1. Params present in request (Ansible) obj are the same as the current (DNAC) params
        # If any does not have eq params, it requires update
        return any(not dnac_compare_equality(current_obj.get(dnac_param),
                                             requested_obj.get(ansible_param))
                   for (dnac_param, ansible_param) in obj_params)

    def create(self):
        result = self.dnac.exec(
            family="discovery",
            function="start_discovery",
            params=self.create_params(),
            op_modifies=True,
        )
        return result

    def update(self):
        id = self.new_object.get("id")
        name = self.new_object.get("name")
        result = None
        if not id:
            prev_obj_name = self.get_object_by_name(name)
            id_ = None
            if prev_obj_name:
                id_ = prev_obj_name.get("id")
            if id_:
                self.new_object.update(dict(id=id_))
        result = self.dnac.exec(
            family="discovery",
            function="updates_discovery_by_id",
            params=self.update_all_params(),
            op_modifies=True,
        )
        return result

    def delete(self):
        id = self.new_object.get("id")
        name = self.new_object.get("name")
        result = None
        if not id:
            prev_obj_name = self.get_object_by_name(name)
            id_ = None
            if prev_obj_name:
                id_ = prev_obj_name.get("id")
            if id_:
                self.new_object.update(dict(id=id_))
        result = self.dnac.exec(
            family="discovery",
            function="delete_discovery_by_id",
            params=self.delete_by_id_params(),
        )
        return result


class ActionModule(ActionBase):
    def __init__(self, *args, **kwargs):
        if not ANSIBLE_UTILS_IS_INSTALLED:
            raise AnsibleActionFail("ansible.utils is not installed. Execute 'ansible-galaxy collection install ansible.utils'")
        super(ActionModule, self).__init__(*args, **kwargs)
        self._supports_async = False
        self._supports_check_mode = False
        self._result = None

    # Checks the supplied parameters against the argument spec for this module
    def _check_argspec(self):
        aav = AnsibleArgSpecValidator(
            data=self._task.args,
            schema=dict(argument_spec=argument_spec),
            schema_format="argspec",
            schema_conditionals=dict(
                required_if=required_if,
                required_one_of=required_one_of,
                mutually_exclusive=mutually_exclusive,
                required_together=required_together,
            ),
            name=self._task.action,
        )
        valid, errors, self._task.args = aav.validate()
        if not valid:
            raise AnsibleActionFail(errors)

    def run(self, tmp=None, task_vars=None):
        self._task.diff = False
        self._result = super(ActionModule, self).run(tmp, task_vars)
        self._result["changed"] = False
        self._check_argspec()

        dnac = DNACSDK(self._task.args)
        obj = Discovery(self._task.args, dnac)

        state = self._task.args.get("state")

        response = None

        if state == "present":
            (obj_exists, prev_obj) = obj.exists()
            if obj_exists:
                if obj.requires_update(prev_obj):
                    response = obj.update()
                    dnac.object_updated()
                else:
                    response = prev_obj
                    dnac.object_already_present()
            else:
                response = obj.create()
                dnac.object_created()

        elif state == "absent":
            (obj_exists, prev_obj) = obj.exists()
            if obj_exists:
                response = obj.delete()
                dnac.object_deleted()
            else:
                dnac.object_already_absent()

        self._result.update(dict(dnac_response=response))
        self._result.update(dnac.exit_json())
        return self._result



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Red Hat Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: rhel_facts
version_added: 1.5.0
short_description: Facts module to set or override RHEL specific facts.
description:
  - Compatibility layer for using the "package" module for rpm-ostree based systems via setting the "pkg_mgr" fact correctly.
author:
  - Adam Miller (@maxamillion)
requirements:
  - rpm-ostree
seealso:
  - module: ansible.builtin.package
options: {}
'''

EXAMPLES = '''
- name: Playbook to use the package module on all RHEL footprints
  vars:
    ansible_facts_modules:
      - setup # REQUIRED to be run before all custom fact modules
      - ansible.posix.rhel_facts
  tasks:
    - name: Ensure packages are installed
      ansible.builtin.package:
        name:
          - htop
          - ansible
        state: present
'''

RETURN = """
ansible_facts:
    description: Relevant Ansible Facts
    returned: when needed
    type: complex
    contains:
        pkg_mgr:
            description: System-level package manager override
            returned: when needed
            type: str
            sample: {'pkg_mgr': 'ansible.posix.rhel_facts'}
"""

import os

from ansible.module_utils.basic import AnsibleModule


def main():

    module = AnsibleModule(
        argument_spec=dict(),
        supports_check_mode=True,
    )

    ansible_facts = {}

    # Verify that the platform is an rpm-ostree based system
    if os.path.exists("/run/ostree-booted"):
        ansible_facts['pkg_mgr'] = 'ansible.posix.rhel_rpm_ostree'

    module.exit_json(ansible_facts=ansible_facts, changed=False)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_snapshot_copy
version_added: 1.0.0
short_description: Copies an EC2 snapshot and returns the new Snapshot ID
description:
  - Copies an EC2 Snapshot from a source region to a destination region.
options:
  source_region:
    description:
      - The source region the Snapshot should be copied from.
    required: true
    type: str
  source_snapshot_id:
    description:
      - The ID of the Snapshot in source region that should be copied.
    required: true
    type: str
  description:
    description:
      - An optional human-readable string describing purpose of the new Snapshot.
    type: str
    default: ''
  encrypted:
    description:
      - Whether or not the destination Snapshot should be encrypted.
    type: bool
    default: false
  kms_key_id:
    description:
      - KMS key id used to encrypt snapshot. If not specified, AWS defaults to C(alias/aws/ebs).
    type: str
  wait:
    description:
      - Wait for the copied Snapshot to be in the C(Available) state before returning.
    type: bool
    default: false
  wait_timeout:
    description:
      - How long before wait gives up, in seconds.
    default: 600
    type: int
  tags:
    description:
      - A dictionary representing the tags to be applied to the newly created resource.
    type: dict
    aliases: ['resource_tags']
author:
  - Deepak Kothandan (@Deepakkothandan) <deepak.kdy@gmail.com>
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
- name: Basic Snapshot Copy
  community.aws.ec2_snapshot_copy:
    source_region: eu-central-1
    region: eu-west-1
    source_snapshot_id: snap-xxxxxxx

- name: Copy Snapshot and wait until available
  community.aws.ec2_snapshot_copy:
    source_region: eu-central-1
    region: eu-west-1
    source_snapshot_id: snap-xxxxxxx
    wait: true
    wait_timeout: 1200   # Default timeout is 600
  register: snapshot_id

- name: Tagged Snapshot copy
  community.aws.ec2_snapshot_copy:
    source_region: eu-central-1
    region: eu-west-1
    source_snapshot_id: snap-xxxxxxx
    tags:
      Name: Snapshot-Name

- name: Encrypted Snapshot copy
  community.aws.ec2_snapshot_copy:
    source_region: eu-central-1
    region: eu-west-1
    source_snapshot_id: snap-xxxxxxx
    encrypted: true

- name: Encrypted Snapshot copy with specified key
  community.aws.ec2_snapshot_copy:
    source_region: eu-central-1
    region: eu-west-1
    source_snapshot_id: snap-xxxxxxx
    encrypted: true
    kms_key_id: arn:aws:kms:eu-central-1:XXXXXXXXXXXX:key/746de6ea-50a4-4bcb-8fbc-e3b29f2d367b
"""

RETURN = r"""
snapshot_id:
    description: snapshot id of the newly created snapshot
    returned: when snapshot copy is successful
    type: str
    sample: "snap-e9095e8c"
"""

try:
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_specifications

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


def copy_snapshot(module, ec2):
    """
    Copies an EC2 Snapshot to another region

    module : AnsibleAWSModule object
    ec2: ec2 connection object
    """

    params = {
        "SourceRegion": module.params.get("source_region"),
        "SourceSnapshotId": module.params.get("source_snapshot_id"),
        "Description": module.params.get("description"),
    }

    if module.params.get("encrypted"):
        params["Encrypted"] = True

    if module.params.get("kms_key_id"):
        params["KmsKeyId"] = module.params.get("kms_key_id")

    if module.params.get("tags"):
        params["TagSpecifications"] = boto3_tag_specifications(module.params.get("tags"), types=["snapshot"])

    try:
        snapshot_id = ec2.copy_snapshot(**params)["SnapshotId"]
        if module.params.get("wait"):
            delay = 15
            # Add one to max_attempts as wait() increment
            # its counter before assessing it for time.sleep()
            max_attempts = (module.params.get("wait_timeout") // delay) + 1
            ec2.get_waiter("snapshot_completed").wait(
                SnapshotIds=[snapshot_id], WaiterConfig=dict(Delay=delay, MaxAttempts=max_attempts)
            )

    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="An error occurred waiting for the snapshot to become available.")

    module.exit_json(changed=True, snapshot_id=snapshot_id)


def main():
    argument_spec = dict(
        source_region=dict(required=True),
        source_snapshot_id=dict(required=True),
        description=dict(default=""),
        encrypted=dict(type="bool", default=False, required=False),
        kms_key_id=dict(type="str", required=False),
        wait=dict(type="bool", default=False),
        wait_timeout=dict(type="int", default=600),
        tags=dict(type="dict", aliases=["resource_tags"]),
    )

    module = AnsibleAWSModule(argument_spec=argument_spec)

    try:
        client = module.client("ec2")
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to connect to AWS")

    copy_snapshot(module, client)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import datetime
import functools
import time
from copy import deepcopy

try:
    import botocore
except ImportError:
    pass  # caught by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict
from ansible.module_utils.six import string_types

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_boto3_tag_list
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.tagging import compare_aws_tags


def get_domain_status(client, module, domain_name):
    """
    Get the status of an existing OpenSearch cluster.
    """
    try:
        response = client.describe_domain(DomainName=domain_name)
    except is_boto3_error_code("ResourceNotFoundException"):
        return None
    except (
        botocore.exceptions.BotoCoreError,
        botocore.exceptions.ClientError,
    ) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg=f"Couldn't get domain {domain_name}")
    return response["DomainStatus"]


def get_domain_config(client, module, domain_name):
    """
    Get the configuration of an existing OpenSearch cluster, convert the data
    such that it can be used as input parameter to client.update_domain().
    The status info is removed.
    The returned config includes the 'EngineVersion' property, it needs to be removed
    from the dict before invoking client.update_domain().

    Return (domain_config, domain_arn) or (None, None) if the domain does not exist.
    """
    try:
        response = client.describe_domain_config(DomainName=domain_name)
    except is_boto3_error_code("ResourceNotFoundException"):
        return (None, None)
    except (
        botocore.exceptions.BotoCoreError,
        botocore.exceptions.ClientError,
    ) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg=f"Couldn't get domain {domain_name}")
    domain_config = {}
    arn = None
    if response is not None:
        for k in response["DomainConfig"]:
            if "Options" in response["DomainConfig"][k]:
                domain_config[k] = response["DomainConfig"][k]["Options"]
        domain_config["DomainName"] = domain_name
        # If ES cluster is attached to the Internet, the "VPCOptions" property is not present.
        if "VPCOptions" in domain_config:
            # The "VPCOptions" returned by the describe_domain_config API has
            # additional attributes that would cause an error if sent in the HTTP POST body.
            dc = {}
            if "SubnetIds" in domain_config["VPCOptions"]:
                dc["SubnetIds"] = deepcopy(domain_config["VPCOptions"]["SubnetIds"])
            if "SecurityGroupIds" in domain_config["VPCOptions"]:
                dc["SecurityGroupIds"] = deepcopy(domain_config["VPCOptions"]["SecurityGroupIds"])
            domain_config["VPCOptions"] = dc
        # The "StartAt" property is converted to datetime, but when doing comparisons it should
        # be in the string format "YYYY-MM-DD".
        for s in domain_config["AutoTuneOptions"]["MaintenanceSchedules"]:
            if isinstance(s["StartAt"], datetime.datetime):
                s["StartAt"] = s["StartAt"].strftime("%Y-%m-%d")
        # Provisioning of "AdvancedOptions" is not supported by this module yet.
        domain_config.pop("AdvancedOptions", None)

        # Get the ARN of the OpenSearch cluster.
        domain = get_domain_status(client, module, domain_name)
        if domain is not None:
            arn = domain["ARN"]
    return (domain_config, arn)


def normalize_opensearch(client, module, domain):
    """
    Merge the input domain object with tags associated with the domain,
    convert the attributes from camel case to snake case, and return the object.
    """
    try:
        domain["Tags"] = boto3_tag_list_to_ansible_dict(client.list_tags(ARN=domain["ARN"], aws_retry=True)["TagList"])
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, f"Couldn't get tags for domain {domain['domain_name']}")
    except KeyError:
        module.fail_json(msg=str(domain))

    return camel_dict_to_snake_dict(domain, ignore_list=["Tags"])


def wait_for_domain_status(client, module, domain_name, waiter_name):
    if not module.params["wait"]:
        return
    timeout = module.params["wait_timeout"]
    deadline = time.time() + timeout
    status_msg = ""
    while time.time() < deadline:
        status = get_domain_status(client, module, domain_name)
        if status is None:
            status_msg = "Not Found"
            if waiter_name == "domain_deleted":
                return
        else:
            status_msg = "Created: {0}. Processing: {1}. UpgradeProcessing: {2}".format(
                status["Created"],
                status["Processing"],
                status["UpgradeProcessing"],
            )
            if (
                waiter_name == "domain_available"
                and status["Created"]
                and not status["Processing"]
                and not status["UpgradeProcessing"]
            ):
                return
        time.sleep(15)
    # Timeout occured.
    module.fail_json(msg=f"Timeout waiting for wait state '{waiter_name}'. {status_msg}")


def parse_version(engine_version):
    """
    Parse the engine version, which should be Elasticsearch_X.Y or OpenSearch_X.Y
    Return dict { 'engine_type': engine_type, 'major': major, 'minor': minor }
    """
    version = engine_version.split("_")
    if len(version) != 2:
        return None
    semver = version[1].split(".")
    if len(semver) != 2:
        return None
    engine_type = version[0]
    if engine_type not in ["Elasticsearch", "OpenSearch"]:
        return None
    if not (semver[0].isdigit() and semver[1].isdigit()):
        return None
    major = int(semver[0])
    minor = int(semver[1])
    return {"engine_type": engine_type, "major": major, "minor": minor}


def compare_domain_versions(version1, version2):
    supported_engines = {
        "Elasticsearch": 1,
        "OpenSearch": 2,
    }
    if isinstance(version1, string_types):
        version1 = parse_version(version1)
    if isinstance(version2, string_types):
        version2 = parse_version(version2)
    if version1 is None and version2 is not None:
        return -1
    elif version1 is not None and version2 is None:
        return 1
    elif version1 is None and version2 is None:
        return 0
    e1 = supported_engines.get(version1.get("engine_type"))
    e2 = supported_engines.get(version2.get("engine_type"))
    if e1 < e2:
        return -1
    elif e1 > e2:
        return 1
    else:
        if version1.get("major") < version2.get("major"):
            return -1
        elif version1.get("major") > version2.get("major"):
            return 1
        else:
            if version1.get("minor") < version2.get("minor"):
                return -1
            elif version1.get("minor") > version2.get("minor"):
                return 1
            else:
                return 0


def get_target_increment_version(client, module, domain_name, target_version):
    """
    Returns the highest compatible version which is less than or equal to target_version.
    When upgrading a domain from version V1 to V2, it may not be possible to upgrade
    directly from V1 to V2. The domain may have to be upgraded through intermediate versions.
    Return None if there is no such version.
    For example, it's not possible to upgrade directly from Elasticsearch 5.5 to 7.10.
    """
    api_compatible_versions = None
    try:
        api_compatible_versions = client.get_compatible_versions(DomainName=domain_name)
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(
            e,
            msg=f"Couldn't get compatible versions for domain {domain_name}",
        )
    compat = api_compatible_versions.get("CompatibleVersions")
    if compat is None:
        module.fail_json("Unable to determine list of compatible versions", compatible_versions=api_compatible_versions)
    if len(compat) == 0:
        module.fail_json("Unable to determine list of compatible versions", compatible_versions=api_compatible_versions)
    if compat[0].get("TargetVersions") is None:
        module.fail_json("No compatible versions found", compatible_versions=api_compatible_versions)
    compatible_versions = []
    for v in compat[0].get("TargetVersions"):
        if target_version == v:
            # It's possible to upgrade directly to the target version.
            return target_version
        semver = parse_version(v)
        if semver is not None:
            compatible_versions.append(semver)
    # No direct upgrade is possible. Upgrade to the highest version available.
    compatible_versions = sorted(compatible_versions, key=functools.cmp_to_key(compare_domain_versions))
    # Return the highest compatible version which is lower than target_version
    for v in reversed(compatible_versions):
        if compare_domain_versions(v, target_version) <= 0:
            return v
    return None


def ensure_tags(client, module, resource_arn, existing_tags, tags, purge_tags):
    if tags is None:
        return False
    tags_to_add, tags_to_remove = compare_aws_tags(existing_tags, tags, purge_tags)
    changed = bool(tags_to_add or tags_to_remove)
    if tags_to_add:
        if module.check_mode:
            module.exit_json(changed=True, msg="Would have added tags to domain if not in check mode")
        try:
            client.add_tags(
                ARN=resource_arn,
                TagList=ansible_dict_to_boto3_tag_list(tags_to_add),
            )
        except (
            botocore.exceptions.ClientError,
            botocore.exceptions.BotoCoreError,
        ) as e:
            module.fail_json_aws(e, f"Couldn't add tags to domain {resource_arn}")
    if tags_to_remove:
        if module.check_mode:
            module.exit_json(changed=True, msg="Would have removed tags if not in check mode")
        try:
            client.remove_tags(ARN=resource_arn, TagKeys=tags_to_remove)
        except (
            botocore.exceptions.ClientError,
            botocore.exceptions.BotoCoreError,
        ) as e:
            module.fail_json_aws(e, f"Couldn't remove tags from domain {resource_arn}")
    return changed
